from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="SandboxStatus")


@_attrs_define
class SandboxStatus:
    """
    Attributes:
        sandbox_id (Union[Unset, str]):
        template_id (Union[Unset, str]):
        team_id (Union[Unset, str]):
        user_id (Union[Unset, str]):
        pod_name (Union[Unset, str]):
        status (Union[Unset, str]):
        claimed_at (Union[Unset, str]):
        expires_at (Union[Unset, str]):
        created_at (Union[Unset, str]):
    """

    sandbox_id: Union[Unset, str] = UNSET
    template_id: Union[Unset, str] = UNSET
    team_id: Union[Unset, str] = UNSET
    user_id: Union[Unset, str] = UNSET
    pod_name: Union[Unset, str] = UNSET
    status: Union[Unset, str] = UNSET
    claimed_at: Union[Unset, str] = UNSET
    expires_at: Union[Unset, str] = UNSET
    created_at: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        sandbox_id = self.sandbox_id

        template_id = self.template_id

        team_id = self.team_id

        user_id = self.user_id

        pod_name = self.pod_name

        status = self.status

        claimed_at = self.claimed_at

        expires_at = self.expires_at

        created_at = self.created_at

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if sandbox_id is not UNSET:
            field_dict["sandbox_id"] = sandbox_id
        if template_id is not UNSET:
            field_dict["template_id"] = template_id
        if team_id is not UNSET:
            field_dict["team_id"] = team_id
        if user_id is not UNSET:
            field_dict["user_id"] = user_id
        if pod_name is not UNSET:
            field_dict["pod_name"] = pod_name
        if status is not UNSET:
            field_dict["status"] = status
        if claimed_at is not UNSET:
            field_dict["claimed_at"] = claimed_at
        if expires_at is not UNSET:
            field_dict["expires_at"] = expires_at
        if created_at is not UNSET:
            field_dict["created_at"] = created_at

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        sandbox_id = d.pop("sandbox_id", UNSET)

        template_id = d.pop("template_id", UNSET)

        team_id = d.pop("team_id", UNSET)

        user_id = d.pop("user_id", UNSET)

        pod_name = d.pop("pod_name", UNSET)

        status = d.pop("status", UNSET)

        claimed_at = d.pop("claimed_at", UNSET)

        expires_at = d.pop("expires_at", UNSET)

        created_at = d.pop("created_at", UNSET)

        sandbox_status = cls(
            sandbox_id=sandbox_id,
            template_id=template_id,
            team_id=team_id,
            user_id=user_id,
            pod_name=pod_name,
            status=status,
            claimed_at=claimed_at,
            expires_at=expires_at,
            created_at=created_at,
        )

        sandbox_status.additional_properties = d
        return sandbox_status

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.repl_config import REPLConfig


T = TypeVar("T", bound="CreateREPLContextRequest")


@_attrs_define
class CreateREPLContextRequest:
    """
    Attributes:
        language (Union[Unset, str]):
        input_ (Union[Unset, str]):
        repl_config (Union[Unset, REPLConfig]):
    """

    language: Union[Unset, str] = UNSET
    input_: Union[Unset, str] = UNSET
    repl_config: Union[Unset, "REPLConfig"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        language = self.language

        input_ = self.input_

        repl_config: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.repl_config, Unset):
            repl_config = self.repl_config.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if language is not UNSET:
            field_dict["language"] = language
        if input_ is not UNSET:
            field_dict["input"] = input_
        if repl_config is not UNSET:
            field_dict["repl_config"] = repl_config

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.repl_config import REPLConfig

        d = dict(src_dict)
        language = d.pop("language", UNSET)

        input_ = d.pop("input", UNSET)

        _repl_config = d.pop("repl_config", UNSET)
        repl_config: Union[Unset, REPLConfig]
        if isinstance(_repl_config, Unset):
            repl_config = UNSET
        else:
            repl_config = REPLConfig.from_dict(_repl_config)

        create_repl_context_request = cls(
            language=language,
            input_=input_,
            repl_config=repl_config,
        )

        create_repl_context_request.additional_properties = d
        return create_repl_context_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.label_selector_match_labels import LabelSelectorMatchLabels
    from ..models.label_selector_requirement import LabelSelectorRequirement


T = TypeVar("T", bound="LabelSelector")


@_attrs_define
class LabelSelector:
    """
    Attributes:
        match_labels (Union[Unset, LabelSelectorMatchLabels]):
        match_expressions (Union[Unset, list['LabelSelectorRequirement']]):
    """

    match_labels: Union[Unset, "LabelSelectorMatchLabels"] = UNSET
    match_expressions: Union[Unset, list["LabelSelectorRequirement"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        match_labels: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.match_labels, Unset):
            match_labels = self.match_labels.to_dict()

        match_expressions: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.match_expressions, Unset):
            match_expressions = []
            for match_expressions_item_data in self.match_expressions:
                match_expressions_item = match_expressions_item_data.to_dict()
                match_expressions.append(match_expressions_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if match_labels is not UNSET:
            field_dict["matchLabels"] = match_labels
        if match_expressions is not UNSET:
            field_dict["matchExpressions"] = match_expressions

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.label_selector_match_labels import LabelSelectorMatchLabels
        from ..models.label_selector_requirement import LabelSelectorRequirement

        d = dict(src_dict)
        _match_labels = d.pop("matchLabels", UNSET)
        match_labels: Union[Unset, LabelSelectorMatchLabels]
        if isinstance(_match_labels, Unset):
            match_labels = UNSET
        else:
            match_labels = LabelSelectorMatchLabels.from_dict(_match_labels)

        match_expressions = []
        _match_expressions = d.pop("matchExpressions", UNSET)
        for match_expressions_item_data in _match_expressions or []:
            match_expressions_item = LabelSelectorRequirement.from_dict(
                match_expressions_item_data
            )

            match_expressions.append(match_expressions_item)

        label_selector = cls(
            match_labels=match_labels,
            match_expressions=match_expressions,
        )

        label_selector.additional_properties = d
        return label_selector

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.sandbox_config import SandboxConfig


T = TypeVar("T", bound="ClaimRequest")


@_attrs_define
class ClaimRequest:
    """
    Attributes:
        template (Union[Unset, str]):
        config (Union[Unset, SandboxConfig]):
    """

    template: Union[Unset, str] = UNSET
    config: Union[Unset, "SandboxConfig"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        template = self.template

        config: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.config, Unset):
            config = self.config.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if template is not UNSET:
            field_dict["template"] = template
        if config is not UNSET:
            field_dict["config"] = config

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.sandbox_config import SandboxConfig

        d = dict(src_dict)
        template = d.pop("template", UNSET)

        _config = d.pop("config", UNSET)
        config: Union[Unset, SandboxConfig]
        if isinstance(_config, Unset):
            config = UNSET
        else:
            config = SandboxConfig.from_dict(_config)

        claim_request = cls(
            template=template,
            config=config,
        )

        claim_request.additional_properties = d
        return claim_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.container_spec import ContainerSpec
    from ..models.lifecycle_policy import LifecyclePolicy
    from ..models.pod_spec_override import PodSpecOverride
    from ..models.pool_strategy import PoolStrategy
    from ..models.sandbox_template_spec_env_vars import SandboxTemplateSpecEnvVars
    from ..models.tpl_sandbox_network_policy import TplSandboxNetworkPolicy


T = TypeVar("T", bound="SandboxTemplateSpec")


@_attrs_define
class SandboxTemplateSpec:
    """
    Attributes:
        description (Union[Unset, str]):
        display_name (Union[Unset, str]):
        tags (Union[Unset, list[str]]):
        main_container (Union[Unset, ContainerSpec]):
        sidecars (Union[Unset, list['ContainerSpec']]):
        pod (Union[Unset, PodSpecOverride]):
        network (Union[Unset, TplSandboxNetworkPolicy]):
        pool (Union[Unset, PoolStrategy]):
        lifecycle (Union[Unset, LifecyclePolicy]):
        env_vars (Union[Unset, SandboxTemplateSpecEnvVars]):
        public (Union[Unset, bool]):
        allowed_teams (Union[Unset, list[str]]):
        runtime_class_name (Union[Unset, str]):
        cluster_id (Union[Unset, str]):
    """

    description: Union[Unset, str] = UNSET
    display_name: Union[Unset, str] = UNSET
    tags: Union[Unset, list[str]] = UNSET
    main_container: Union[Unset, "ContainerSpec"] = UNSET
    sidecars: Union[Unset, list["ContainerSpec"]] = UNSET
    pod: Union[Unset, "PodSpecOverride"] = UNSET
    network: Union[Unset, "TplSandboxNetworkPolicy"] = UNSET
    pool: Union[Unset, "PoolStrategy"] = UNSET
    lifecycle: Union[Unset, "LifecyclePolicy"] = UNSET
    env_vars: Union[Unset, "SandboxTemplateSpecEnvVars"] = UNSET
    public: Union[Unset, bool] = UNSET
    allowed_teams: Union[Unset, list[str]] = UNSET
    runtime_class_name: Union[Unset, str] = UNSET
    cluster_id: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        description = self.description

        display_name = self.display_name

        tags: Union[Unset, list[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        main_container: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.main_container, Unset):
            main_container = self.main_container.to_dict()

        sidecars: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.sidecars, Unset):
            sidecars = []
            for sidecars_item_data in self.sidecars:
                sidecars_item = sidecars_item_data.to_dict()
                sidecars.append(sidecars_item)

        pod: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.pod, Unset):
            pod = self.pod.to_dict()

        network: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.network, Unset):
            network = self.network.to_dict()

        pool: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.pool, Unset):
            pool = self.pool.to_dict()

        lifecycle: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.lifecycle, Unset):
            lifecycle = self.lifecycle.to_dict()

        env_vars: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.env_vars, Unset):
            env_vars = self.env_vars.to_dict()

        public = self.public

        allowed_teams: Union[Unset, list[str]] = UNSET
        if not isinstance(self.allowed_teams, Unset):
            allowed_teams = self.allowed_teams

        runtime_class_name = self.runtime_class_name

        cluster_id = self.cluster_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if description is not UNSET:
            field_dict["description"] = description
        if display_name is not UNSET:
            field_dict["displayName"] = display_name
        if tags is not UNSET:
            field_dict["tags"] = tags
        if main_container is not UNSET:
            field_dict["mainContainer"] = main_container
        if sidecars is not UNSET:
            field_dict["sidecars"] = sidecars
        if pod is not UNSET:
            field_dict["pod"] = pod
        if network is not UNSET:
            field_dict["network"] = network
        if pool is not UNSET:
            field_dict["pool"] = pool
        if lifecycle is not UNSET:
            field_dict["lifecycle"] = lifecycle
        if env_vars is not UNSET:
            field_dict["envVars"] = env_vars
        if public is not UNSET:
            field_dict["public"] = public
        if allowed_teams is not UNSET:
            field_dict["allowedTeams"] = allowed_teams
        if runtime_class_name is not UNSET:
            field_dict["runtimeClassName"] = runtime_class_name
        if cluster_id is not UNSET:
            field_dict["clusterId"] = cluster_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.container_spec import ContainerSpec
        from ..models.lifecycle_policy import LifecyclePolicy
        from ..models.pod_spec_override import PodSpecOverride
        from ..models.pool_strategy import PoolStrategy
        from ..models.sandbox_template_spec_env_vars import SandboxTemplateSpecEnvVars
        from ..models.tpl_sandbox_network_policy import TplSandboxNetworkPolicy

        d = dict(src_dict)
        description = d.pop("description", UNSET)

        display_name = d.pop("displayName", UNSET)

        tags = cast(list[str], d.pop("tags", UNSET))

        _main_container = d.pop("mainContainer", UNSET)
        main_container: Union[Unset, ContainerSpec]
        if isinstance(_main_container, Unset):
            main_container = UNSET
        else:
            main_container = ContainerSpec.from_dict(_main_container)

        sidecars = []
        _sidecars = d.pop("sidecars", UNSET)
        for sidecars_item_data in _sidecars or []:
            sidecars_item = ContainerSpec.from_dict(sidecars_item_data)

            sidecars.append(sidecars_item)

        _pod = d.pop("pod", UNSET)
        pod: Union[Unset, PodSpecOverride]
        if isinstance(_pod, Unset):
            pod = UNSET
        else:
            pod = PodSpecOverride.from_dict(_pod)

        _network = d.pop("network", UNSET)
        network: Union[Unset, TplSandboxNetworkPolicy]
        if isinstance(_network, Unset):
            network = UNSET
        else:
            network = TplSandboxNetworkPolicy.from_dict(_network)

        _pool = d.pop("pool", UNSET)
        pool: Union[Unset, PoolStrategy]
        if isinstance(_pool, Unset):
            pool = UNSET
        else:
            pool = PoolStrategy.from_dict(_pool)

        _lifecycle = d.pop("lifecycle", UNSET)
        lifecycle: Union[Unset, LifecyclePolicy]
        if isinstance(_lifecycle, Unset):
            lifecycle = UNSET
        else:
            lifecycle = LifecyclePolicy.from_dict(_lifecycle)

        _env_vars = d.pop("envVars", UNSET)
        env_vars: Union[Unset, SandboxTemplateSpecEnvVars]
        if isinstance(_env_vars, Unset):
            env_vars = UNSET
        else:
            env_vars = SandboxTemplateSpecEnvVars.from_dict(_env_vars)

        public = d.pop("public", UNSET)

        allowed_teams = cast(list[str], d.pop("allowedTeams", UNSET))

        runtime_class_name = d.pop("runtimeClassName", UNSET)

        cluster_id = d.pop("clusterId", UNSET)

        sandbox_template_spec = cls(
            description=description,
            display_name=display_name,
            tags=tags,
            main_container=main_container,
            sidecars=sidecars,
            pod=pod,
            network=network,
            pool=pool,
            lifecycle=lifecycle,
            env_vars=env_vars,
            public=public,
            allowed_teams=allowed_teams,
            runtime_class_name=runtime_class_name,
            cluster_id=cluster_id,
        )

        sandbox_template_spec.additional_properties = d
        return sandbox_template_spec

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import base64
import json
from io import BytesIO
from typing import TYPE_CHECKING, Any

from sandbox0.apispec.api.files import delete_api_v1_sandboxes_id_files
from sandbox0.apispec.api.files import get_api_v1_sandboxes_id_files
from sandbox0.apispec.api.files import get_api_v1_sandboxes_id_files_list
from sandbox0.apispec.api.files import get_api_v1_sandboxes_id_files_stat
from sandbox0.apispec.api.files import post_api_v1_sandboxes_id_files
from sandbox0.apispec.api.files import post_api_v1_sandboxes_id_files_move
from sandbox0.apispec.models.file_content_response import FileContentResponse
from sandbox0.apispec.models.file_content_response_encoding import FileContentResponseEncoding
from sandbox0.apispec.models.file_info import FileInfo
from sandbox0.apispec.models.move_file_request import MoveFileRequest
from sandbox0.apispec.models.success_created_response import SuccessCreatedResponse
from sandbox0.apispec.models.success_deleted_response import SuccessDeletedResponse
from sandbox0.apispec.models.success_file_list_response import SuccessFileListResponse
from sandbox0.apispec.models.success_file_read_response import SuccessFileReadResponse
from sandbox0.apispec.models.success_file_stat_response import SuccessFileStatResponse
from sandbox0.apispec.models.success_moved_response import SuccessMovedResponse
from sandbox0.apispec.models.success_written_response import SuccessWrittenResponse
from sandbox0.apispec.types import File as APIFile
from sandbox0.apispec.types import UNSET
from sandbox0.errors import APIError
from sandbox0.models import FileWatchResponse
from sandbox0.response import ensure_data, ensure_model

if TYPE_CHECKING:
    from sandbox0.sandbox import Sandbox
    from websockets.sync.client import ClientConnection


class FileWatchStream:
    def __init__(self, conn: "ClientConnection", watch_id: str) -> None:
        self._conn = conn
        self.watch_id = watch_id

    def iter_events(self) -> Any:
        while True:
            raw = self._conn.recv()
            if raw is None:
                return
            if isinstance(raw, bytes):
                raw = raw.decode("utf-8", errors="replace")
            payload = json.loads(raw)
            yield FileWatchResponse(
                type=str(payload.get("type", "")),
                watch_id=str(payload.get("watch_id", "")),
                event=str(payload.get("event", "")),
                path=str(payload.get("path", "")),
                error=str(payload.get("error", "")),
            )

    def unsubscribe(self) -> None:
        self._conn.send(json.dumps({"action": "unsubscribe", "watch_id": self.watch_id}))
        self._conn.close()

    def close(self) -> None:
        self._conn.close()


class SandboxFilesMixin:
    id: str
    _client: any

    def read_file(self: "Sandbox", path: str) -> bytes:
        resp = get_api_v1_sandboxes_id_files.sync_detailed(id=self.id, client=self._client.api, path=path)

        content_type = str(resp.headers.get("Content-Type", "")).lower()
        if "application/json" in content_type:
            envelope = ensure_model(resp, SuccessFileReadResponse)
            data = envelope.data
            if data.__class__.__name__ == "Unset":
                raise APIError(status_code=int(resp.status_code), message="unexpected empty file response")
            if isinstance(data, FileContentResponse):
                if data.encoding.__class__.__name__ != "Unset" and data.encoding != FileContentResponseEncoding.BASE64:
                    raise APIError(status_code=int(resp.status_code), message=f"unsupported file encoding: {data.encoding}")
                content = "" if data.content.__class__.__name__ == "Unset" else data.content
                return base64.b64decode(content)
        if resp.parsed is None:
            raise APIError(status_code=int(resp.status_code), message="unexpected empty file response")
        return resp.content

    def stat_file(self: "Sandbox", path: str) -> FileInfo:
        resp = get_api_v1_sandboxes_id_files_stat.sync_detailed(id=self.id, client=self._client.api, path=path)
        return ensure_data(resp, SuccessFileStatResponse)

    def list_files(self: "Sandbox", path: str) -> list[FileInfo]:
        resp = get_api_v1_sandboxes_id_files_list.sync_detailed(id=self.id, client=self._client.api, path=path)
        data = ensure_data(resp, SuccessFileListResponse)
        entries = data.entries
        if entries.__class__.__name__ == "Unset":
            return []
        return entries

    def write_file(self: "Sandbox", path: str, data: bytes) -> SuccessWrittenResponse:
        resp = post_api_v1_sandboxes_id_files.sync_detailed(
            id=self.id,
            client=self._client.api,
            path=path,
            body=APIFile(payload=BytesIO(data)),
        )
        parsed = resp.parsed
        if isinstance(parsed, SuccessWrittenResponse):
            return parsed
        if isinstance(parsed, SuccessCreatedResponse):
            raise APIError(status_code=int(resp.status_code), message="directory created instead of file")
        raise APIError(status_code=int(resp.status_code), message="unexpected response")

    def mkdir(self: "Sandbox", path: str, recursive: bool = False) -> SuccessCreatedResponse:
        resp = post_api_v1_sandboxes_id_files.sync_detailed(
            id=self.id,
            client=self._client.api,
            path=path,
            mkdir=True,
            recursive=recursive if recursive else UNSET,
            body=APIFile(payload=BytesIO(b"")),
        )
        return ensure_model(resp, SuccessCreatedResponse)

    def delete_file(self: "Sandbox", path: str) -> SuccessDeletedResponse:
        resp = delete_api_v1_sandboxes_id_files.sync_detailed(id=self.id, client=self._client.api, path=path)
        return ensure_model(resp, SuccessDeletedResponse)

    def move_file(self: "Sandbox", source: str, destination: str) -> SuccessMovedResponse:
        resp = post_api_v1_sandboxes_id_files_move.sync_detailed(
            id=self.id,
            client=self._client.api,
            body=MoveFileRequest(source=source, destination=destination),
        )
        return ensure_model(resp, SuccessMovedResponse)

    def watch_files(self: "Sandbox", path: str, recursive: bool = False) -> FileWatchStream:
        from websockets.sync.client import connect

        ws_url = self._client.websocket_url(f"/api/v1/sandboxes/{self.id}/files/watch")
        conn = connect(ws_url, additional_headers=self._client.ws_headers())
        conn.send(json.dumps({"action": "subscribe", "path": path, "recursive": recursive}))
        raw = conn.recv()
        if raw is None:
            conn.close()
            raise APIError(status_code=0, message="watch subscribe failed: empty response")
        if isinstance(raw, bytes):
            raw = raw.decode("utf-8", errors="replace")
        payload = json.loads(raw)
        if payload.get("type") == "error":
            conn.close()
            raise APIError(status_code=0, message=f"watch subscribe failed: {payload.get('error', '')}")
        if payload.get("type") != "subscribed" or not payload.get("watch_id"):
            conn.close()
            raise APIError(status_code=0, message=f"unexpected watch response: {payload.get('type', '')}")
        return FileWatchStream(conn=conn, watch_id=str(payload["watch_id"]))



import datetime
from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

T = TypeVar("T", bound="Team")


@_attrs_define
class Team:
    """
    Attributes:
        id (str):
        name (str):
        slug (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        owner_id (Union[None, Unset, str]):
    """

    id: str
    name: str
    slug: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    owner_id: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        name = self.name

        slug = self.slug

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        owner_id: Union[None, Unset, str]
        if isinstance(self.owner_id, Unset):
            owner_id = UNSET
        else:
            owner_id = self.owner_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "name": name,
                "slug": slug,
                "created_at": created_at,
                "updated_at": updated_at,
            }
        )
        if owner_id is not UNSET:
            field_dict["owner_id"] = owner_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        id = d.pop("id")

        name = d.pop("name")

        slug = d.pop("slug")

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        def _parse_owner_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        owner_id = _parse_owner_id(d.pop("owner_id", UNSET))

        team = cls(
            id=id,
            name=name,
            slug=slug,
            created_at=created_at,
            updated_at=updated_at,
            owner_id=owner_id,
        )

        team.additional_properties = d
        return team

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

from typing import TYPE_CHECKING

from sandbox0.apispec.api.templates import delete_api_v1_templates_id
from sandbox0.apispec.api.templates import get_api_v1_templates
from sandbox0.apispec.api.templates import get_api_v1_templates_id
from sandbox0.apispec.api.templates import post_api_v1_templates
from sandbox0.apispec.api.templates import put_api_v1_templates_id
from sandbox0.apispec.models.success_message_response import SuccessMessageResponse
from sandbox0.apispec.models.success_template_list_response import SuccessTemplateListResponse
from sandbox0.apispec.models.success_template_response import SuccessTemplateResponse
from sandbox0.apispec.models.template import Template
from sandbox0.apispec.models.template_create_request import TemplateCreateRequest
from sandbox0.apispec.models.template_update_request import TemplateUpdateRequest
from sandbox0.response import ensure_data, ensure_model

if TYPE_CHECKING:
    from sandbox0.client import Client


class ClientTemplatesMixin:
    _api: any

    def list_templates(self: "Client") -> list[Template]:
        resp = get_api_v1_templates.sync_detailed(client=self._api)
        data = ensure_data(resp, SuccessTemplateListResponse)
        templates = data.templates
        if templates.__class__.__name__ == "Unset":
            return []
        return templates

    # Backward-compatible alias.
    def list_template(self: "Client") -> list[Template]:
        return self.list_templates()

    def get_template(self: "Client", template_id: str) -> Template:
        resp = get_api_v1_templates_id.sync_detailed(id=template_id, client=self._api)
        return ensure_data(resp, SuccessTemplateResponse)

    def create_template(self: "Client", request: TemplateCreateRequest) -> Template:
        resp = post_api_v1_templates.sync_detailed(client=self._api, body=request)
        return ensure_data(resp, SuccessTemplateResponse)

    def update_template(self: "Client", template_id: str, request: TemplateUpdateRequest) -> Template:
        resp = put_api_v1_templates_id.sync_detailed(id=template_id, client=self._api, body=request)
        return ensure_data(resp, SuccessTemplateResponse)

    def delete_template(self: "Client", template_id: str) -> SuccessMessageResponse:
        resp = delete_api_v1_templates_id.sync_detailed(id=template_id, client=self._api)
        return ensure_model(resp, SuccessMessageResponse)



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.resource_usage import ResourceUsage


T = TypeVar("T", bound="ContextResourceUsage")


@_attrs_define
class ContextResourceUsage:
    """
    Attributes:
        context_id (Union[Unset, str]):
        type_ (Union[Unset, str]):
        language (Union[Unset, str]):
        running (Union[Unset, bool]):
        paused (Union[Unset, bool]):
        usage (Union[Unset, ResourceUsage]):
    """

    context_id: Union[Unset, str] = UNSET
    type_: Union[Unset, str] = UNSET
    language: Union[Unset, str] = UNSET
    running: Union[Unset, bool] = UNSET
    paused: Union[Unset, bool] = UNSET
    usage: Union[Unset, "ResourceUsage"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        context_id = self.context_id

        type_ = self.type_

        language = self.language

        running = self.running

        paused = self.paused

        usage: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.usage, Unset):
            usage = self.usage.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if context_id is not UNSET:
            field_dict["context_id"] = context_id
        if type_ is not UNSET:
            field_dict["type"] = type_
        if language is not UNSET:
            field_dict["language"] = language
        if running is not UNSET:
            field_dict["running"] = running
        if paused is not UNSET:
            field_dict["paused"] = paused
        if usage is not UNSET:
            field_dict["usage"] = usage

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.resource_usage import ResourceUsage

        d = dict(src_dict)
        context_id = d.pop("context_id", UNSET)

        type_ = d.pop("type", UNSET)

        language = d.pop("language", UNSET)

        running = d.pop("running", UNSET)

        paused = d.pop("paused", UNSET)

        _usage = d.pop("usage", UNSET)
        usage: Union[Unset, ResourceUsage]
        if isinstance(_usage, Unset):
            usage = UNSET
        else:
            usage = ResourceUsage.from_dict(_usage)

        context_resource_usage = cls(
            context_id=context_id,
            type_=type_,
            language=language,
            running=running,
            paused=paused,
            usage=usage,
        )

        context_resource_usage.additional_properties = d
        return context_resource_usage

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.exposed_port_config import ExposedPortConfig
    from ..models.sandbox_config_env_vars import SandboxConfigEnvVars
    from ..models.tpl_sandbox_network_policy import TplSandboxNetworkPolicy
    from ..models.webhook_config import WebhookConfig


T = TypeVar("T", bound="SandboxConfig")


@_attrs_define
class SandboxConfig:
    """
    Attributes:
        env_vars (Union[Unset, SandboxConfigEnvVars]):
        ttl (Union[Unset, int]):
        hard_ttl (Union[Unset, int]):
        network (Union[Unset, TplSandboxNetworkPolicy]):
        webhook (Union[Unset, WebhookConfig]):
        auto_resume (Union[Unset, bool]): Sandbox-level resume gate for paused sandboxes. When false, any inbound
            request
            (API or public exposure) must not auto resume the sandbox.
             Default: False.
        exposed_ports (Union[Unset, list['ExposedPortConfig']]):
    """

    env_vars: Union[Unset, "SandboxConfigEnvVars"] = UNSET
    ttl: Union[Unset, int] = UNSET
    hard_ttl: Union[Unset, int] = UNSET
    network: Union[Unset, "TplSandboxNetworkPolicy"] = UNSET
    webhook: Union[Unset, "WebhookConfig"] = UNSET
    auto_resume: Union[Unset, bool] = False
    exposed_ports: Union[Unset, list["ExposedPortConfig"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        env_vars: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.env_vars, Unset):
            env_vars = self.env_vars.to_dict()

        ttl = self.ttl

        hard_ttl = self.hard_ttl

        network: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.network, Unset):
            network = self.network.to_dict()

        webhook: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.webhook, Unset):
            webhook = self.webhook.to_dict()

        auto_resume = self.auto_resume

        exposed_ports: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.exposed_ports, Unset):
            exposed_ports = []
            for exposed_ports_item_data in self.exposed_ports:
                exposed_ports_item = exposed_ports_item_data.to_dict()
                exposed_ports.append(exposed_ports_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if env_vars is not UNSET:
            field_dict["env_vars"] = env_vars
        if ttl is not UNSET:
            field_dict["ttl"] = ttl
        if hard_ttl is not UNSET:
            field_dict["hard_ttl"] = hard_ttl
        if network is not UNSET:
            field_dict["network"] = network
        if webhook is not UNSET:
            field_dict["webhook"] = webhook
        if auto_resume is not UNSET:
            field_dict["auto_resume"] = auto_resume
        if exposed_ports is not UNSET:
            field_dict["exposed_ports"] = exposed_ports

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.exposed_port_config import ExposedPortConfig
        from ..models.sandbox_config_env_vars import SandboxConfigEnvVars
        from ..models.tpl_sandbox_network_policy import TplSandboxNetworkPolicy
        from ..models.webhook_config import WebhookConfig

        d = dict(src_dict)
        _env_vars = d.pop("env_vars", UNSET)
        env_vars: Union[Unset, SandboxConfigEnvVars]
        if isinstance(_env_vars, Unset):
            env_vars = UNSET
        else:
            env_vars = SandboxConfigEnvVars.from_dict(_env_vars)

        ttl = d.pop("ttl", UNSET)

        hard_ttl = d.pop("hard_ttl", UNSET)

        _network = d.pop("network", UNSET)
        network: Union[Unset, TplSandboxNetworkPolicy]
        if isinstance(_network, Unset):
            network = UNSET
        else:
            network = TplSandboxNetworkPolicy.from_dict(_network)

        _webhook = d.pop("webhook", UNSET)
        webhook: Union[Unset, WebhookConfig]
        if isinstance(_webhook, Unset):
            webhook = UNSET
        else:
            webhook = WebhookConfig.from_dict(_webhook)

        auto_resume = d.pop("auto_resume", UNSET)

        exposed_ports = []
        _exposed_ports = d.pop("exposed_ports", UNSET)
        for exposed_ports_item_data in _exposed_ports or []:
            exposed_ports_item = ExposedPortConfig.from_dict(exposed_ports_item_data)

            exposed_ports.append(exposed_ports_item)

        sandbox_config = cls(
            env_vars=env_vars,
            ttl=ttl,
            hard_ttl=hard_ttl,
            network=network,
            webhook=webhook,
            auto_resume=auto_resume,
            exposed_ports=exposed_ports,
        )

        sandbox_config.additional_properties = d
        return sandbox_config

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.node_affinity import NodeAffinity
    from ..models.pod_affinity import PodAffinity


T = TypeVar("T", bound="Affinity")


@_attrs_define
class Affinity:
    """
    Attributes:
        node_affinity (Union[Unset, NodeAffinity]):
        pod_affinity (Union[Unset, PodAffinity]):
    """

    node_affinity: Union[Unset, "NodeAffinity"] = UNSET
    pod_affinity: Union[Unset, "PodAffinity"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        node_affinity: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.node_affinity, Unset):
            node_affinity = self.node_affinity.to_dict()

        pod_affinity: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.pod_affinity, Unset):
            pod_affinity = self.pod_affinity.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if node_affinity is not UNSET:
            field_dict["nodeAffinity"] = node_affinity
        if pod_affinity is not UNSET:
            field_dict["podAffinity"] = pod_affinity

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.node_affinity import NodeAffinity
        from ..models.pod_affinity import PodAffinity

        d = dict(src_dict)
        _node_affinity = d.pop("nodeAffinity", UNSET)
        node_affinity: Union[Unset, NodeAffinity]
        if isinstance(_node_affinity, Unset):
            node_affinity = UNSET
        else:
            node_affinity = NodeAffinity.from_dict(_node_affinity)

        _pod_affinity = d.pop("podAffinity", UNSET)
        pod_affinity: Union[Unset, PodAffinity]
        if isinstance(_pod_affinity, Unset):
            pod_affinity = UNSET
        else:
            pod_affinity = PodAffinity.from_dict(_pod_affinity)

        affinity = cls(
            node_affinity=node_affinity,
            pod_affinity=pod_affinity,
        )

        affinity.additional_properties = d
        return affinity

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class UpdateTeamMemberRequestRole(str, Enum):
    ADMIN = "admin"
    DEVELOPER = "developer"
    VIEWER = "viewer"

    def __str__(self) -> str:
        return str(self.value)



from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="ExposedPortConfig")


@_attrs_define
class ExposedPortConfig:
    """
    Attributes:
        port (int):
        resume (bool): Port-level resume gate for public exposure traffic. Evaluated only when
            sandbox auto_resume is true. Priority: sandbox auto_resume (global gate)
            > exposed_ports[].resume (per-port gate).
        public_url (Union[Unset, str]): The full public URL to access this exposed port.
            Format: https://<sandboxName>--p<port>.<regionID>.<rootDomain>
    """

    port: int
    resume: bool
    public_url: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        port = self.port

        resume = self.resume

        public_url = self.public_url

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "port": port,
                "resume": resume,
            }
        )
        if public_url is not UNSET:
            field_dict["public_url"] = public_url

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        port = d.pop("port")

        resume = d.pop("resume")

        public_url = d.pop("public_url", UNSET)

        exposed_port_config = cls(
            port=port,
            resume=resume,
            public_url=public_url,
        )

        exposed_port_config.additional_properties = d
        return exposed_port_config

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.exec_candidate import ExecCandidate
    from ..models.repl_env_var import REPLEnvVar
    from ..models.repl_prompt_config import REPLPromptConfig
    from ..models.repl_ready_config import REPLReadyConfig


T = TypeVar("T", bound="REPLConfig")


@_attrs_define
class REPLConfig:
    """
    Attributes:
        name (str):
        candidates (list['ExecCandidate']):
        display_name (Union[Unset, str]):
        description (Union[Unset, str]):
        env (Union[Unset, list['REPLEnvVar']]):
        default_term (Union[Unset, str]):
        prompt (Union[Unset, REPLPromptConfig]):
        ready (Union[Unset, REPLReadyConfig]):
    """

    name: str
    candidates: list["ExecCandidate"]
    display_name: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    env: Union[Unset, list["REPLEnvVar"]] = UNSET
    default_term: Union[Unset, str] = UNSET
    prompt: Union[Unset, "REPLPromptConfig"] = UNSET
    ready: Union[Unset, "REPLReadyConfig"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        name = self.name

        candidates = []
        for candidates_item_data in self.candidates:
            candidates_item = candidates_item_data.to_dict()
            candidates.append(candidates_item)

        display_name = self.display_name

        description = self.description

        env: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.env, Unset):
            env = []
            for env_item_data in self.env:
                env_item = env_item_data.to_dict()
                env.append(env_item)

        default_term = self.default_term

        prompt: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.prompt, Unset):
            prompt = self.prompt.to_dict()

        ready: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.ready, Unset):
            ready = self.ready.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "candidates": candidates,
            }
        )
        if display_name is not UNSET:
            field_dict["display_name"] = display_name
        if description is not UNSET:
            field_dict["description"] = description
        if env is not UNSET:
            field_dict["env"] = env
        if default_term is not UNSET:
            field_dict["default_term"] = default_term
        if prompt is not UNSET:
            field_dict["prompt"] = prompt
        if ready is not UNSET:
            field_dict["ready"] = ready

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.exec_candidate import ExecCandidate
        from ..models.repl_env_var import REPLEnvVar
        from ..models.repl_prompt_config import REPLPromptConfig
        from ..models.repl_ready_config import REPLReadyConfig

        d = dict(src_dict)
        name = d.pop("name")

        candidates = []
        _candidates = d.pop("candidates")
        for candidates_item_data in _candidates:
            candidates_item = ExecCandidate.from_dict(candidates_item_data)

            candidates.append(candidates_item)

        display_name = d.pop("display_name", UNSET)

        description = d.pop("description", UNSET)

        env = []
        _env = d.pop("env", UNSET)
        for env_item_data in _env or []:
            env_item = REPLEnvVar.from_dict(env_item_data)

            env.append(env_item)

        default_term = d.pop("default_term", UNSET)

        _prompt = d.pop("prompt", UNSET)
        prompt: Union[Unset, REPLPromptConfig]
        if isinstance(_prompt, Unset):
            prompt = UNSET
        else:
            prompt = REPLPromptConfig.from_dict(_prompt)

        _ready = d.pop("ready", UNSET)
        ready: Union[Unset, REPLReadyConfig]
        if isinstance(_ready, Unset):
            ready = UNSET
        else:
            ready = REPLReadyConfig.from_dict(_ready)

        repl_config = cls(
            name=name,
            candidates=candidates,
            display_name=display_name,
            description=description,
            env=env,
            default_term=default_term,
            prompt=prompt,
            ready=ready,
        )

        repl_config.additional_properties = d
        return repl_config

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.object_meta import ObjectMeta
    from ..models.sandbox_template_spec import SandboxTemplateSpec
    from ..models.sandbox_template_status import SandboxTemplateStatus


T = TypeVar("T", bound="SandboxTemplate")


@_attrs_define
class SandboxTemplate:
    """
    Attributes:
        api_version (Union[Unset, str]):
        kind (Union[Unset, str]):
        metadata (Union[Unset, ObjectMeta]):
        spec (Union[Unset, SandboxTemplateSpec]):
        status (Union[Unset, SandboxTemplateStatus]):
    """

    api_version: Union[Unset, str] = UNSET
    kind: Union[Unset, str] = UNSET
    metadata: Union[Unset, "ObjectMeta"] = UNSET
    spec: Union[Unset, "SandboxTemplateSpec"] = UNSET
    status: Union[Unset, "SandboxTemplateStatus"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        api_version = self.api_version

        kind = self.kind

        metadata: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.metadata, Unset):
            metadata = self.metadata.to_dict()

        spec: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.spec, Unset):
            spec = self.spec.to_dict()

        status: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if api_version is not UNSET:
            field_dict["apiVersion"] = api_version
        if kind is not UNSET:
            field_dict["kind"] = kind
        if metadata is not UNSET:
            field_dict["metadata"] = metadata
        if spec is not UNSET:
            field_dict["spec"] = spec
        if status is not UNSET:
            field_dict["status"] = status

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.object_meta import ObjectMeta
        from ..models.sandbox_template_spec import SandboxTemplateSpec
        from ..models.sandbox_template_status import SandboxTemplateStatus

        d = dict(src_dict)
        api_version = d.pop("apiVersion", UNSET)

        kind = d.pop("kind", UNSET)

        _metadata = d.pop("metadata", UNSET)
        metadata: Union[Unset, ObjectMeta]
        if isinstance(_metadata, Unset):
            metadata = UNSET
        else:
            metadata = ObjectMeta.from_dict(_metadata)

        _spec = d.pop("spec", UNSET)
        spec: Union[Unset, SandboxTemplateSpec]
        if isinstance(_spec, Unset):
            spec = UNSET
        else:
            spec = SandboxTemplateSpec.from_dict(_spec)

        _status = d.pop("status", UNSET)
        status: Union[Unset, SandboxTemplateStatus]
        if isinstance(_status, Unset):
            status = UNSET
        else:
            status = SandboxTemplateStatus.from_dict(_status)

        sandbox_template = cls(
            api_version=api_version,
            kind=kind,
            metadata=metadata,
            spec=spec,
            status=status,
        )

        sandbox_template.additional_properties = d
        return sandbox_template

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="ResourceUsage")


@_attrs_define
class ResourceUsage:
    """
    Attributes:
        cpu_percent (Union[Unset, float]):
        memory_rss (Union[Unset, int]):
        memory_vms (Union[Unset, int]):
        open_files (Union[Unset, int]):
        thread_count (Union[Unset, int]):
        container_memory_usage (Union[Unset, int]):
        container_memory_limit (Union[Unset, int]):
        container_memory_working_set (Union[Unset, int]):
        io_read_bytes (Union[Unset, int]):
        io_write_bytes (Union[Unset, int]):
        memory_bytes (Union[Unset, int]):
    """

    cpu_percent: Union[Unset, float] = UNSET
    memory_rss: Union[Unset, int] = UNSET
    memory_vms: Union[Unset, int] = UNSET
    open_files: Union[Unset, int] = UNSET
    thread_count: Union[Unset, int] = UNSET
    container_memory_usage: Union[Unset, int] = UNSET
    container_memory_limit: Union[Unset, int] = UNSET
    container_memory_working_set: Union[Unset, int] = UNSET
    io_read_bytes: Union[Unset, int] = UNSET
    io_write_bytes: Union[Unset, int] = UNSET
    memory_bytes: Union[Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        cpu_percent = self.cpu_percent

        memory_rss = self.memory_rss

        memory_vms = self.memory_vms

        open_files = self.open_files

        thread_count = self.thread_count

        container_memory_usage = self.container_memory_usage

        container_memory_limit = self.container_memory_limit

        container_memory_working_set = self.container_memory_working_set

        io_read_bytes = self.io_read_bytes

        io_write_bytes = self.io_write_bytes

        memory_bytes = self.memory_bytes

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if cpu_percent is not UNSET:
            field_dict["cpu_percent"] = cpu_percent
        if memory_rss is not UNSET:
            field_dict["memory_rss"] = memory_rss
        if memory_vms is not UNSET:
            field_dict["memory_vms"] = memory_vms
        if open_files is not UNSET:
            field_dict["open_files"] = open_files
        if thread_count is not UNSET:
            field_dict["thread_count"] = thread_count
        if container_memory_usage is not UNSET:
            field_dict["container_memory_usage"] = container_memory_usage
        if container_memory_limit is not UNSET:
            field_dict["container_memory_limit"] = container_memory_limit
        if container_memory_working_set is not UNSET:
            field_dict["container_memory_working_set"] = container_memory_working_set
        if io_read_bytes is not UNSET:
            field_dict["io_read_bytes"] = io_read_bytes
        if io_write_bytes is not UNSET:
            field_dict["io_write_bytes"] = io_write_bytes
        if memory_bytes is not UNSET:
            field_dict["memory_bytes"] = memory_bytes

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        cpu_percent = d.pop("cpu_percent", UNSET)

        memory_rss = d.pop("memory_rss", UNSET)

        memory_vms = d.pop("memory_vms", UNSET)

        open_files = d.pop("open_files", UNSET)

        thread_count = d.pop("thread_count", UNSET)

        container_memory_usage = d.pop("container_memory_usage", UNSET)

        container_memory_limit = d.pop("container_memory_limit", UNSET)

        container_memory_working_set = d.pop("container_memory_working_set", UNSET)

        io_read_bytes = d.pop("io_read_bytes", UNSET)

        io_write_bytes = d.pop("io_write_bytes", UNSET)

        memory_bytes = d.pop("memory_bytes", UNSET)

        resource_usage = cls(
            cpu_percent=cpu_percent,
            memory_rss=memory_rss,
            memory_vms=memory_vms,
            open_files=open_files,
            thread_count=thread_count,
            container_memory_usage=container_memory_usage,
            container_memory_limit=container_memory_limit,
            container_memory_working_set=container_memory_working_set,
            io_read_bytes=io_read_bytes,
            io_write_bytes=io_write_bytes,
            memory_bytes=memory_bytes,
        )

        resource_usage.additional_properties = d
        return resource_usage

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

from typing import Dict, Optional
from urllib.parse import urlparse, urlunparse

import httpx

from sandbox0.apispec.client import AuthenticatedClient
from sandbox0.resources import Sandboxes, Volumes
from sandbox0.response_normalize import normalize_response_hook, normalize_response_hook_async
from sandbox0.sandbox import Sandbox

DEFAULT_BASE_URL = "https://api.sandbox0.ai"


class Client:
    def __init__(
        self,
        *,
        token: str,
        base_url: str = DEFAULT_BASE_URL,
        timeout: Optional[float] = None,
        user_agent: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        verify_ssl: bool = True,
        raise_on_unexpected_status: bool = False,
    ) -> None:
        merged_headers: Dict[str, str] = {}
        if headers:
            merged_headers.update(headers)
        if user_agent:
            merged_headers["User-Agent"] = user_agent

        httpx_timeout = httpx.Timeout(timeout) if timeout is not None else None
        self._api = AuthenticatedClient(
            base_url=base_url,
            token=token,
            headers=merged_headers,
            timeout=httpx_timeout,
            verify_ssl=verify_ssl,
            raise_on_unexpected_status=raise_on_unexpected_status,
        )
        auth_value = f"{self._api.prefix} {self._api.token}" if self._api.prefix else self._api.token
        request_headers = dict(self._api._headers)
        request_headers[self._api.auth_header_name] = auth_value
        self._api.set_httpx_client(
            httpx.Client(
                base_url=base_url,
                headers=request_headers,
                cookies=self._api._cookies,
                timeout=httpx_timeout,
                verify=verify_ssl,
                follow_redirects=self._api._follow_redirects,
                event_hooks={"response": [normalize_response_hook]},
            )
        )
        self._api.set_async_httpx_client(
            httpx.AsyncClient(
                base_url=base_url,
                headers=request_headers,
                cookies=self._api._cookies,
                timeout=httpx_timeout,
                verify=verify_ssl,
                follow_redirects=self._api._follow_redirects,
                event_hooks={"response": [normalize_response_hook_async]},
            )
        )
        self.base_url = base_url
        self.sandboxes = Sandboxes(self)
        self.volumes = Volumes(self)

    @property
    def api(self) -> AuthenticatedClient:
        return self._api

    def sandbox(self, sandbox_id: str) -> Sandbox:
        return Sandbox(id=sandbox_id, client=self)

    def websocket_url(self, path: str) -> str:
        parsed = urlparse(self.base_url)
        scheme = parsed.scheme.lower()
        if scheme == "https":
            ws_scheme = "wss"
        elif scheme == "http":
            ws_scheme = "ws"
        elif scheme in ("ws", "wss"):
            ws_scheme = scheme
        else:
            raise ValueError(f"unsupported URL scheme: {parsed.scheme}")
        normalized_path = path if path.startswith("/") else "/" + path
        return urlunparse((ws_scheme, parsed.netloc, normalized_path, "", "", ""))

    def ws_headers(self) -> dict[str, str]:
        # Reuse generated client's auth/header behavior for WebSocket upgrade requests.
        http_client = self._api.get_httpx_client()
        return dict(http_client.headers)



from __future__ import annotations

import json
from typing import Any, Optional, Type, TypeVar

from sandbox0.apispec.models.error_envelope import ErrorEnvelope
from sandbox0.apispec.types import Response

from sandbox0.errors import APIError

T = TypeVar("T")


def ensure_model(response: Response[Any], model_type: Type[T]) -> T:
    parsed = response.parsed
    if isinstance(parsed, ErrorEnvelope):
        raise _to_api_error(response, parsed)
    if isinstance(parsed, model_type):
        return parsed
    raise _unexpected_response(response)


def ensure_data(response: Response[Any], envelope_type: Type[Any]) -> Any:
    parsed = response.parsed
    if isinstance(parsed, ErrorEnvelope):
        raise _to_api_error(response, parsed)
    if isinstance(parsed, envelope_type):
        data = getattr(parsed, "data", None)
        if _is_unset(data) or data is None:
            raise _unexpected_response(response)
        return data
    raise _unexpected_response(response)


def optional_data(response: Response[Any], envelope_type: Type[Any]) -> Any:
    parsed = response.parsed
    if isinstance(parsed, ErrorEnvelope):
        raise _to_api_error(response, parsed)
    if isinstance(parsed, envelope_type):
        return getattr(parsed, "data", None)
    raise _unexpected_response(response)


def _is_unset(value: Any) -> bool:
    return value is None or value.__class__.__name__ == "Unset"


def _to_api_error(response: Response[Any], envelope: ErrorEnvelope) -> APIError:
    return APIError(
        status_code=int(response.status_code),
        code=envelope.error.code,
        message=envelope.error.message,
        details=envelope.error.details,
        request_id=_request_id_from_headers(response.headers),
        body=response.content,
    )


def _unexpected_response(response: Response[Any]) -> APIError:
    body = response.content or b""
    message = "Unexpected API response"
    if body:
        try:
            payload = json.loads(body.decode("utf-8"))
            message = f"{message}: {json.dumps(payload, separators=(',', ':'), ensure_ascii=True)}"
        except Exception:
            message = f"{message}: {body.decode('utf-8', errors='replace').strip()}"
    return APIError(
        status_code=int(response.status_code),
        message=message,
        request_id=_request_id_from_headers(response.headers),
        body=response.content,
    )


def _request_id_from_headers(headers: dict[str, str]) -> Optional[str]:
    for key in ("X-Request-Id", "X-Request-ID", "Request-Id", "Request-ID"):
        value = headers.get(key)
        if value:
            stripped = value.strip()
            if stripped:
                return stripped
    return None



from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="MountStatus")


@_attrs_define
class MountStatus:
    """
    Attributes:
        sandboxvolume_id (Union[Unset, str]):
        mount_point (Union[Unset, str]):
        mounted_at (Union[Unset, str]):
        mounted_duration_sec (Union[Unset, int]):
        mount_session_id (Union[Unset, str]):
    """

    sandboxvolume_id: Union[Unset, str] = UNSET
    mount_point: Union[Unset, str] = UNSET
    mounted_at: Union[Unset, str] = UNSET
    mounted_duration_sec: Union[Unset, int] = UNSET
    mount_session_id: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        sandboxvolume_id = self.sandboxvolume_id

        mount_point = self.mount_point

        mounted_at = self.mounted_at

        mounted_duration_sec = self.mounted_duration_sec

        mount_session_id = self.mount_session_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if sandboxvolume_id is not UNSET:
            field_dict["sandboxvolume_id"] = sandboxvolume_id
        if mount_point is not UNSET:
            field_dict["mount_point"] = mount_point
        if mounted_at is not UNSET:
            field_dict["mounted_at"] = mounted_at
        if mounted_duration_sec is not UNSET:
            field_dict["mounted_duration_sec"] = mounted_duration_sec
        if mount_session_id is not UNSET:
            field_dict["mount_session_id"] = mount_session_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        sandboxvolume_id = d.pop("sandboxvolume_id", UNSET)

        mount_point = d.pop("mount_point", UNSET)

        mounted_at = d.pop("mounted_at", UNSET)

        mounted_duration_sec = d.pop("mounted_duration_sec", UNSET)

        mount_session_id = d.pop("mount_session_id", UNSET)

        mount_status = cls(
            sandboxvolume_id=sandboxvolume_id,
            mount_point=mount_point,
            mounted_at=mounted_at,
            mounted_duration_sec=mounted_duration_sec,
            mount_session_id=mount_session_id,
        )

        mount_status.additional_properties = d
        return mount_status

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.context_resource_usage import ContextResourceUsage


T = TypeVar("T", bound="SandboxResourceUsage")


@_attrs_define
class SandboxResourceUsage:
    """
    Attributes:
        container_memory_usage (Union[Unset, int]):
        container_memory_limit (Union[Unset, int]):
        container_memory_working_set (Union[Unset, int]):
        total_memory_rss (Union[Unset, int]):
        total_memory_vms (Union[Unset, int]):
        total_open_files (Union[Unset, int]):
        total_thread_count (Union[Unset, int]):
        total_io_read_bytes (Union[Unset, int]):
        total_io_write_bytes (Union[Unset, int]):
        context_count (Union[Unset, int]):
        running_context_count (Union[Unset, int]):
        paused_context_count (Union[Unset, int]):
        contexts (Union[Unset, list['ContextResourceUsage']]):
    """

    container_memory_usage: Union[Unset, int] = UNSET
    container_memory_limit: Union[Unset, int] = UNSET
    container_memory_working_set: Union[Unset, int] = UNSET
    total_memory_rss: Union[Unset, int] = UNSET
    total_memory_vms: Union[Unset, int] = UNSET
    total_open_files: Union[Unset, int] = UNSET
    total_thread_count: Union[Unset, int] = UNSET
    total_io_read_bytes: Union[Unset, int] = UNSET
    total_io_write_bytes: Union[Unset, int] = UNSET
    context_count: Union[Unset, int] = UNSET
    running_context_count: Union[Unset, int] = UNSET
    paused_context_count: Union[Unset, int] = UNSET
    contexts: Union[Unset, list["ContextResourceUsage"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        container_memory_usage = self.container_memory_usage

        container_memory_limit = self.container_memory_limit

        container_memory_working_set = self.container_memory_working_set

        total_memory_rss = self.total_memory_rss

        total_memory_vms = self.total_memory_vms

        total_open_files = self.total_open_files

        total_thread_count = self.total_thread_count

        total_io_read_bytes = self.total_io_read_bytes

        total_io_write_bytes = self.total_io_write_bytes

        context_count = self.context_count

        running_context_count = self.running_context_count

        paused_context_count = self.paused_context_count

        contexts: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.contexts, Unset):
            contexts = []
            for contexts_item_data in self.contexts:
                contexts_item = contexts_item_data.to_dict()
                contexts.append(contexts_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if container_memory_usage is not UNSET:
            field_dict["container_memory_usage"] = container_memory_usage
        if container_memory_limit is not UNSET:
            field_dict["container_memory_limit"] = container_memory_limit
        if container_memory_working_set is not UNSET:
            field_dict["container_memory_working_set"] = container_memory_working_set
        if total_memory_rss is not UNSET:
            field_dict["total_memory_rss"] = total_memory_rss
        if total_memory_vms is not UNSET:
            field_dict["total_memory_vms"] = total_memory_vms
        if total_open_files is not UNSET:
            field_dict["total_open_files"] = total_open_files
        if total_thread_count is not UNSET:
            field_dict["total_thread_count"] = total_thread_count
        if total_io_read_bytes is not UNSET:
            field_dict["total_io_read_bytes"] = total_io_read_bytes
        if total_io_write_bytes is not UNSET:
            field_dict["total_io_write_bytes"] = total_io_write_bytes
        if context_count is not UNSET:
            field_dict["context_count"] = context_count
        if running_context_count is not UNSET:
            field_dict["running_context_count"] = running_context_count
        if paused_context_count is not UNSET:
            field_dict["paused_context_count"] = paused_context_count
        if contexts is not UNSET:
            field_dict["contexts"] = contexts

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.context_resource_usage import ContextResourceUsage

        d = dict(src_dict)
        container_memory_usage = d.pop("container_memory_usage", UNSET)

        container_memory_limit = d.pop("container_memory_limit", UNSET)

        container_memory_working_set = d.pop("container_memory_working_set", UNSET)

        total_memory_rss = d.pop("total_memory_rss", UNSET)

        total_memory_vms = d.pop("total_memory_vms", UNSET)

        total_open_files = d.pop("total_open_files", UNSET)

        total_thread_count = d.pop("total_thread_count", UNSET)

        total_io_read_bytes = d.pop("total_io_read_bytes", UNSET)

        total_io_write_bytes = d.pop("total_io_write_bytes", UNSET)

        context_count = d.pop("context_count", UNSET)

        running_context_count = d.pop("running_context_count", UNSET)

        paused_context_count = d.pop("paused_context_count", UNSET)

        contexts = []
        _contexts = d.pop("contexts", UNSET)
        for contexts_item_data in _contexts or []:
            contexts_item = ContextResourceUsage.from_dict(contexts_item_data)

            contexts.append(contexts_item)

        sandbox_resource_usage = cls(
            container_memory_usage=container_memory_usage,
            container_memory_limit=container_memory_limit,
            container_memory_working_set=container_memory_working_set,
            total_memory_rss=total_memory_rss,
            total_memory_vms=total_memory_vms,
            total_open_files=total_open_files,
            total_thread_count=total_thread_count,
            total_io_read_bytes=total_io_read_bytes,
            total_io_write_bytes=total_io_write_bytes,
            context_count=context_count,
            running_context_count=running_context_count,
            paused_context_count=paused_context_count,
            contexts=contexts,
        )

        sandbox_resource_usage.additional_properties = d
        return sandbox_resource_usage

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

from typing import TYPE_CHECKING

from sandbox0.apispec.api.sandboxes import delete_api_v1_sandboxes_id_exposed_ports
from sandbox0.apispec.api.sandboxes import delete_api_v1_sandboxes_id_exposed_ports_port
from sandbox0.apispec.api.sandboxes import get_api_v1_sandboxes_id_exposed_ports
from sandbox0.apispec.api.sandboxes import put_api_v1_sandboxes_id_exposed_ports
from sandbox0.apispec.models.exposed_port_config import ExposedPortConfig
from sandbox0.apispec.models.success_exposed_ports_response import SuccessExposedPortsResponse
from sandbox0.apispec.models.update_exposed_ports_request import UpdateExposedPortsRequest
from sandbox0.models import ExposedPort, ExposedPortsResponse
from sandbox0.response import ensure_data

if TYPE_CHECKING:
    from sandbox0.sandbox import Sandbox


class SandboxExposedPortsMixin:
    id: str
    _client: any

    def get_exposed_ports(self: "Sandbox") -> ExposedPortsResponse:
        resp = get_api_v1_sandboxes_id_exposed_ports.sync_detailed(id=self.id, client=self._client.api)
        data = ensure_data(resp, SuccessExposedPortsResponse)
        return _to_exposed_ports_response(data)

    def update_exposed_ports(self: "Sandbox", ports: list[ExposedPort]) -> ExposedPortsResponse:
        request = UpdateExposedPortsRequest(
            ports=[ExposedPortConfig(port=p.port, resume=p.resume) for p in ports]
        )
        resp = put_api_v1_sandboxes_id_exposed_ports.sync_detailed(id=self.id, client=self._client.api, body=request)
        data = ensure_data(resp, SuccessExposedPortsResponse)
        return _to_exposed_ports_response(data)

    def expose_port(self: "Sandbox", port: int, resume: bool) -> ExposedPortsResponse:
        current = self.get_exposed_ports()
        ports = list(current.ports)
        updated = False
        for index, existing in enumerate(ports):
            if existing.port == port:
                ports[index] = ExposedPort(port=port, resume=resume, public_url=existing.public_url)
                updated = True
                break
        if not updated:
            ports.append(ExposedPort(port=port, resume=resume))
        return self.update_exposed_ports(ports)

    def unexpose_port(self: "Sandbox", port: int) -> ExposedPortsResponse:
        resp = delete_api_v1_sandboxes_id_exposed_ports_port.sync_detailed(
            id=self.id,
            port=port,
            client=self._client.api,
        )
        data = ensure_data(resp, SuccessExposedPortsResponse)
        return _to_exposed_ports_response(data)

    def clear_exposed_ports(self: "Sandbox") -> None:
        delete_api_v1_sandboxes_id_exposed_ports.sync_detailed(id=self.id, client=self._client.api)


def _to_exposed_ports_response(data: any) -> ExposedPortsResponse:
    ports: list[ExposedPort] = []
    for item in data.exposed_ports:
        public_url = "" if item.public_url.__class__.__name__ == "Unset" else str(item.public_url)
        ports.append(ExposedPort(port=item.port, resume=item.resume, public_url=public_url))
    domain = "" if data.exposure_domain.__class__.__name__ == "Unset" else str(data.exposure_domain)
    return ExposedPortsResponse(ports=ports, exposure_domain=domain)



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.sandbox_resource_usage import SandboxResourceUsage


T = TypeVar("T", bound="PauseSandboxResponse")


@_attrs_define
class PauseSandboxResponse:
    """
    Attributes:
        sandbox_id (str):
        paused (bool):
        resource_usage (Union[Unset, SandboxResourceUsage]):
        updated_memory (Union[Unset, str]):
        updated_cpu (Union[Unset, str]):
    """

    sandbox_id: str
    paused: bool
    resource_usage: Union[Unset, "SandboxResourceUsage"] = UNSET
    updated_memory: Union[Unset, str] = UNSET
    updated_cpu: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        sandbox_id = self.sandbox_id

        paused = self.paused

        resource_usage: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.resource_usage, Unset):
            resource_usage = self.resource_usage.to_dict()

        updated_memory = self.updated_memory

        updated_cpu = self.updated_cpu

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "sandbox_id": sandbox_id,
                "paused": paused,
            }
        )
        if resource_usage is not UNSET:
            field_dict["resource_usage"] = resource_usage
        if updated_memory is not UNSET:
            field_dict["updated_memory"] = updated_memory
        if updated_cpu is not UNSET:
            field_dict["updated_cpu"] = updated_cpu

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.sandbox_resource_usage import SandboxResourceUsage

        d = dict(src_dict)
        sandbox_id = d.pop("sandbox_id")

        paused = d.pop("paused")

        _resource_usage = d.pop("resource_usage", UNSET)
        resource_usage: Union[Unset, SandboxResourceUsage]
        if isinstance(_resource_usage, Unset):
            resource_usage = UNSET
        else:
            resource_usage = SandboxResourceUsage.from_dict(_resource_usage)

        updated_memory = d.pop("updated_memory", UNSET)

        updated_cpu = d.pop("updated_cpu", UNSET)

        pause_sandbox_response = cls(
            sandbox_id=sandbox_id,
            paused=paused,
            resource_usage=resource_usage,
            updated_memory=updated_memory,
            updated_cpu=updated_cpu,
        )

        pause_sandbox_response.additional_properties = d
        return pause_sandbox_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.user import User


T = TypeVar("T", bound="LoginResponse")


@_attrs_define
class LoginResponse:
    """
    Attributes:
        access_token (str):
        refresh_token (str):
        expires_at (int):
        user (User):
    """

    access_token: str
    refresh_token: str
    expires_at: int
    user: "User"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        access_token = self.access_token

        refresh_token = self.refresh_token

        expires_at = self.expires_at

        user = self.user.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "access_token": access_token,
                "refresh_token": refresh_token,
                "expires_at": expires_at,
                "user": user,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.user import User

        d = dict(src_dict)
        access_token = d.pop("access_token")

        refresh_token = d.pop("refresh_token")

        expires_at = d.pop("expires_at")

        user = User.from_dict(d.pop("user"))

        login_response = cls(
            access_token=access_token,
            refresh_token=refresh_token,
            expires_at=expires_at,
            user=user,
        )

        login_response.additional_properties = d
        return login_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.sandbox_template_condition import SandboxTemplateCondition


T = TypeVar("T", bound="SandboxTemplateStatus")


@_attrs_define
class SandboxTemplateStatus:
    """
    Attributes:
        idle_count (Union[Unset, int]):
        active_count (Union[Unset, int]):
        conditions (Union[Unset, list['SandboxTemplateCondition']]):
        last_update_time (Union[Unset, datetime.datetime]):
    """

    idle_count: Union[Unset, int] = UNSET
    active_count: Union[Unset, int] = UNSET
    conditions: Union[Unset, list["SandboxTemplateCondition"]] = UNSET
    last_update_time: Union[Unset, datetime.datetime] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        idle_count = self.idle_count

        active_count = self.active_count

        conditions: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.conditions, Unset):
            conditions = []
            for conditions_item_data in self.conditions:
                conditions_item = conditions_item_data.to_dict()
                conditions.append(conditions_item)

        last_update_time: Union[Unset, str] = UNSET
        if not isinstance(self.last_update_time, Unset):
            last_update_time = self.last_update_time.isoformat()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if idle_count is not UNSET:
            field_dict["idleCount"] = idle_count
        if active_count is not UNSET:
            field_dict["activeCount"] = active_count
        if conditions is not UNSET:
            field_dict["conditions"] = conditions
        if last_update_time is not UNSET:
            field_dict["lastUpdateTime"] = last_update_time

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.sandbox_template_condition import SandboxTemplateCondition

        d = dict(src_dict)
        idle_count = d.pop("idleCount", UNSET)

        active_count = d.pop("activeCount", UNSET)

        conditions = []
        _conditions = d.pop("conditions", UNSET)
        for conditions_item_data in _conditions or []:
            conditions_item = SandboxTemplateCondition.from_dict(conditions_item_data)

            conditions.append(conditions_item)

        _last_update_time = d.pop("lastUpdateTime", UNSET)
        last_update_time: Union[Unset, datetime.datetime]
        if isinstance(_last_update_time, Unset):
            last_update_time = UNSET
        else:
            last_update_time = isoparse(_last_update_time)

        sandbox_template_status = cls(
            idle_count=idle_count,
            active_count=active_count,
            conditions=conditions,
            last_update_time=last_update_time,
        )

        sandbox_template_status.additional_properties = d
        return sandbox_template_status

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import json
import shlex
import threading
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Optional

from sandbox0.apispec.models.create_cmd_context_request import CreateCMDContextRequest
from sandbox0.apispec.models.create_context_request import CreateContextRequest
from sandbox0.apispec.models.create_context_request_env_vars import CreateContextRequestEnvVars
from sandbox0.apispec.models.create_repl_context_request import CreateREPLContextRequest
from sandbox0.apispec.models.pty_size import PTYSize
from sandbox0.apispec.models.process_type import ProcessType
from sandbox0.apispec.types import UNSET
from sandbox0.errors import APIError
from sandbox0.models import CmdResult, RunResult, StreamInput, StreamOutput
from sandbox0.sandbox_contexts import SandboxContextsMixin
from sandbox0.sandbox_exposed_ports import SandboxExposedPortsMixin
from sandbox0.sandbox_files import SandboxFilesMixin
from sandbox0.sandbox_network import SandboxNetworkMixin
from sandbox0.sandbox_volumes import SandboxVolumesMixin

if TYPE_CHECKING:
    from websockets.sync.client import ClientConnection


@dataclass
class RunOptions:
    context_id: Optional[str] = None
    idle_timeout_sec: Optional[int] = None
    ttl_sec: Optional[int] = None
    cwd: Optional[str] = None
    env_vars: Optional[Dict[str, str]] = None
    pty_rows: Optional[int] = None
    pty_cols: Optional[int] = None


@dataclass
class CmdOptions:
    command: Optional[list[str]] = None
    wait: Optional[bool] = None
    idle_timeout_sec: Optional[int] = None
    ttl_sec: Optional[int] = None
    cwd: Optional[str] = None
    env_vars: Optional[Dict[str, str]] = None
    pty_rows: Optional[int] = None
    pty_cols: Optional[int] = None


class ContextStream:
    def __init__(self, conn: "ClientConnection", sandbox_id: str, context_id: str) -> None:
        self._conn = conn
        self._sandbox_id = sandbox_id
        self._context_id = context_id
        self._send_lock = threading.Lock()

    @property
    def context_id(self) -> str:
        return self._context_id

    def send(self, message: StreamInput) -> None:
        payload: dict[str, Any] = {"type": message.type}
        if message.data:
            payload["data"] = message.data
        if message.rows > 0:
            payload["rows"] = message.rows
        if message.cols > 0:
            payload["cols"] = message.cols
        if message.signal:
            payload["signal"] = message.signal
        if message.request_id:
            payload["request_id"] = message.request_id
        with self._send_lock:
            self._conn.send(json.dumps(payload))

    def send_input(self, data: str, request_id: Optional[str] = None) -> None:
        self.send(
            StreamInput(
                type="input",
                data=data,
                request_id=request_id or f"req-{time.time_ns()}",
            )
        )

    def send_resize(self, rows: int, cols: int) -> None:
        self.send(StreamInput(type="resize", rows=rows, cols=cols))

    def send_signal(self, signal: str) -> None:
        self.send(StreamInput(type="signal", signal=signal))

    def iter_outputs(self) -> Any:
        from websockets.exceptions import ConnectionClosedOK

        while True:
            try:
                raw = self._conn.recv()
            except ConnectionClosedOK:
                return
            if raw is None:
                return
            if isinstance(raw, bytes):
                raw = raw.decode("utf-8", errors="replace")
            payload = json.loads(raw)
            yield StreamOutput(
                sandbox_id=self._sandbox_id,
                context_id=self._context_id,
                source=str(payload.get("source", "")),
                data=str(payload.get("data", "")),
            )

    def close(self) -> None:
        self._conn.close()


class Sandbox(
    SandboxContextsMixin,
    SandboxFilesMixin,
    SandboxNetworkMixin,
    SandboxVolumesMixin,
    SandboxExposedPortsMixin,
):
    def __init__(
        self,
        *,
        id: str,
        client: Any,
        template: str = "",
        cluster_id: Optional[str] = None,
        pod_name: str = "",
        status: str = "",
    ) -> None:
        self.id = id
        self.template = template
        self.cluster_id = cluster_id
        self.pod_name = pod_name
        self.status = status
        self._client = client
        self._repl_context_by_lang: dict[str, str] = {}
        self._lock = threading.Lock()

    def run(self, language: str, input: str, options: Optional[RunOptions] = None) -> RunResult:
        if not input.strip():
            raise ValueError("input cannot be empty")
        opts = options or RunOptions()
        context_id = self._ensure_repl_context(language=language, options=opts)
        exec_resp = self.context_exec(context_id=context_id, input=input)
        return RunResult(sandbox_id=self.id, context_id=context_id, output_raw=exec_resp.output_raw)

    def cmd(self, cmd: str, options: Optional[CmdOptions] = None) -> CmdResult:
        if not cmd.strip():
            raise ValueError("command cannot be empty")
        opts = options or CmdOptions()
        command = opts.command if opts.command is not None else shlex.split(cmd)
        if not command:
            raise ValueError("command cannot be empty")
        wait = True if opts.wait is None else opts.wait
        request = CreateContextRequest(
            type_=ProcessType.CMD,
            cmd=CreateCMDContextRequest(command=command),
            wait_until_done=wait,
            cwd=opts.cwd if opts.cwd is not None else UNSET,
            env_vars=self._build_env_vars(opts.env_vars),
            pty_size=self._build_pty(opts.pty_rows, opts.pty_cols),
            idle_timeout_sec=opts.idle_timeout_sec if opts.idle_timeout_sec is not None else UNSET,
            ttl_sec=opts.ttl_sec if opts.ttl_sec is not None else UNSET,
        )
        context_resp = self.create_context(request=request)
        output_raw = "" if context_resp.output_raw.__class__.__name__ == "Unset" else context_resp.output_raw
        return CmdResult(sandbox_id=self.id, context_id=context_resp.id, output_raw=output_raw)

    def run_stream(
        self,
        language: str,
        options: Optional[RunOptions] = None,
    ) -> ContextStream:
        opts = options or RunOptions()
        context_id = self._ensure_repl_context(language=language, options=opts)
        return self.connect_ws_context(context_id=context_id)

    def cmd_stream(self, cmd: str, options: Optional[CmdOptions] = None) -> ContextStream:
        if not cmd.strip():
            raise ValueError("command cannot be empty")
        opts = options or CmdOptions()
        command = opts.command if opts.command is not None else shlex.split(cmd)
        if not command:
            raise ValueError("command cannot be empty")
        wait = False if opts.wait is None else opts.wait
        request = CreateContextRequest(
            type_=ProcessType.CMD,
            cmd=CreateCMDContextRequest(command=command),
            wait_until_done=wait,
            cwd=opts.cwd if opts.cwd is not None else UNSET,
            env_vars=self._build_env_vars(opts.env_vars),
            pty_size=self._build_pty(opts.pty_rows, opts.pty_cols),
            idle_timeout_sec=opts.idle_timeout_sec if opts.idle_timeout_sec is not None else UNSET,
            ttl_sec=opts.ttl_sec if opts.ttl_sec is not None else UNSET,
        )
        context_resp = self.create_context(request=request)
        return self.connect_ws_context(context_id=context_resp.id)

    def connect_ws_context(self, context_id: str) -> ContextStream:
        from websockets.sync.client import connect

        ws_url = self._client.websocket_url(f"/api/v1/sandboxes/{self.id}/contexts/{context_id}/ws")
        conn = connect(ws_url, additional_headers=self._client.ws_headers())
        return ContextStream(conn=conn, sandbox_id=self.id, context_id=context_id)

    def _ensure_repl_context(self, language: str, options: RunOptions) -> str:
        if options.context_id:
            return options.context_id
        lang = language.strip() or "python"
        with self._lock:
            cached = self._repl_context_by_lang.get(lang)
        if cached:
            return cached
        request = CreateContextRequest(
            type_=ProcessType.REPL,
            repl=CreateREPLContextRequest(language=lang),
            cwd=options.cwd if options.cwd is not None else UNSET,
            env_vars=self._build_env_vars(options.env_vars),
            pty_size=self._build_pty(options.pty_rows, options.pty_cols),
            idle_timeout_sec=options.idle_timeout_sec if options.idle_timeout_sec is not None else UNSET,
            ttl_sec=options.ttl_sec if options.ttl_sec is not None else UNSET,
        )
        context_resp = self.create_context(request=request)
        context_id = context_resp.id
        with self._lock:
            self._repl_context_by_lang[lang] = context_id
        return context_id

    @staticmethod
    def _build_pty(rows: Optional[int], cols: Optional[int]) -> Any:
        if rows is None or cols is None:
            return UNSET
        if rows <= 0 or cols <= 0:
            raise APIError(status_code=0, message="pty rows and cols must be > 0")
        return PTYSize(rows=rows, cols=cols)

    @staticmethod
    def _build_env_vars(env_vars: Optional[Dict[str, str]]) -> Any:
        if env_vars is None:
            return UNSET
        model = CreateContextRequestEnvVars()
        model.additional_properties = dict(env_vars)
        return model




import datetime
from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.exposed_port_config import ExposedPortConfig


T = TypeVar("T", bound="Sandbox")


@_attrs_define
class Sandbox:
    """
    Attributes:
        id (str):
        template_id (str):
        team_id (str):
        status (str):
        paused (bool):
        auto_resume (bool):
        pod_name (str):
        expires_at (datetime.datetime):
        claimed_at (datetime.datetime):
        created_at (datetime.datetime):
        user_id (Union[Unset, str]):
        exposed_ports (Union[Unset, list['ExposedPortConfig']]):
    """

    id: str
    template_id: str
    team_id: str
    status: str
    paused: bool
    auto_resume: bool
    pod_name: str
    expires_at: datetime.datetime
    claimed_at: datetime.datetime
    created_at: datetime.datetime
    user_id: Union[Unset, str] = UNSET
    exposed_ports: Union[Unset, list["ExposedPortConfig"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        template_id = self.template_id

        team_id = self.team_id

        status = self.status

        paused = self.paused

        auto_resume = self.auto_resume

        pod_name = self.pod_name

        expires_at = self.expires_at.isoformat()

        claimed_at = self.claimed_at.isoformat()

        created_at = self.created_at.isoformat()

        user_id = self.user_id

        exposed_ports: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.exposed_ports, Unset):
            exposed_ports = []
            for exposed_ports_item_data in self.exposed_ports:
                exposed_ports_item = exposed_ports_item_data.to_dict()
                exposed_ports.append(exposed_ports_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "template_id": template_id,
                "team_id": team_id,
                "status": status,
                "paused": paused,
                "auto_resume": auto_resume,
                "pod_name": pod_name,
                "expires_at": expires_at,
                "claimed_at": claimed_at,
                "created_at": created_at,
            }
        )
        if user_id is not UNSET:
            field_dict["user_id"] = user_id
        if exposed_ports is not UNSET:
            field_dict["exposed_ports"] = exposed_ports

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.exposed_port_config import ExposedPortConfig

        d = dict(src_dict)
        id = d.pop("id")

        template_id = d.pop("template_id")

        team_id = d.pop("team_id")

        status = d.pop("status")

        paused = d.pop("paused")

        auto_resume = d.pop("auto_resume")

        pod_name = d.pop("pod_name")

        expires_at = isoparse(d.pop("expires_at"))

        claimed_at = isoparse(d.pop("claimed_at"))

        created_at = isoparse(d.pop("created_at"))

        user_id = d.pop("user_id", UNSET)

        exposed_ports = []
        _exposed_ports = d.pop("exposed_ports", UNSET)
        for exposed_ports_item_data in _exposed_ports or []:
            exposed_ports_item = ExposedPortConfig.from_dict(exposed_ports_item_data)

            exposed_ports.append(exposed_ports_item)

        sandbox = cls(
            id=id,
            template_id=template_id,
            team_id=team_id,
            status=status,
            paused=paused,
            auto_resume=auto_resume,
            pod_name=pod_name,
            expires_at=expires_at,
            claimed_at=claimed_at,
            created_at=created_at,
            user_id=user_id,
            exposed_ports=exposed_ports,
        )

        sandbox.additional_properties = d
        return sandbox

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

from typing import Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from sandbox0.sandbox import Sandbox
    from sandbox0.apispec.models.sandbox_volume import SandboxVolume


class SandboxSession:
    def __init__(self, sandbox: "Sandbox", closer: Callable[[], None]) -> None:
        self._sandbox = sandbox
        self._closer = closer
        self._closed = False

    @property
    def sandbox(self) -> "Sandbox":
        return self._sandbox

    def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        self._closer()

    def __enter__(self) -> "Sandbox":
        return self._sandbox

    def __exit__(self, exc_type, exc, tb) -> bool:
        self.close()
        return False


class VolumeSession:
    def __init__(self, volume: "SandboxVolume", closer: Callable[[], None]) -> None:
        self._volume = volume
        self._closer = closer
        self._closed = False

    @property
    def volume(self) -> "SandboxVolume":
        return self._volume

    def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        self._closer()

    def __enter__(self) -> "SandboxVolume":
        return self._volume

    def __exit__(self, exc_type, exc, tb) -> bool:
        self.close()
        return False


class MountSession:
    def __init__(
        self,
        *,
        volume_id: str,
        mount_point: str,
        mount_session_id: str,
        unmount: Callable[[], None],
    ) -> None:
        self.volume_id = volume_id
        self.mount_point = mount_point
        self.mount_session_id = mount_session_id
        self._unmount = unmount
        self._closed = False

    def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        self._unmount()

    def __enter__(self) -> "MountSession":
        return self

    def __exit__(self, exc_type, exc, tb) -> bool:
        self.close()
        return False



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.volume_config import VolumeConfig


T = TypeVar("T", bound="MountRequest")


@_attrs_define
class MountRequest:
    """
    Attributes:
        sandboxvolume_id (str):
        mount_point (str):
        sandbox_id (Union[Unset, str]):
        volume_config (Union[Unset, VolumeConfig]):
    """

    sandboxvolume_id: str
    mount_point: str
    sandbox_id: Union[Unset, str] = UNSET
    volume_config: Union[Unset, "VolumeConfig"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        sandboxvolume_id = self.sandboxvolume_id

        mount_point = self.mount_point

        sandbox_id = self.sandbox_id

        volume_config: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.volume_config, Unset):
            volume_config = self.volume_config.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "sandboxvolume_id": sandboxvolume_id,
                "mount_point": mount_point,
            }
        )
        if sandbox_id is not UNSET:
            field_dict["sandbox_id"] = sandbox_id
        if volume_config is not UNSET:
            field_dict["volume_config"] = volume_config

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.volume_config import VolumeConfig

        d = dict(src_dict)
        sandboxvolume_id = d.pop("sandboxvolume_id")

        mount_point = d.pop("mount_point")

        sandbox_id = d.pop("sandbox_id", UNSET)

        _volume_config = d.pop("volume_config", UNSET)
        volume_config: Union[Unset, VolumeConfig]
        if isinstance(_volume_config, Unset):
            volume_config = UNSET
        else:
            volume_config = VolumeConfig.from_dict(_volume_config)

        mount_request = cls(
            sandboxvolume_id=sandboxvolume_id,
            mount_point=mount_point,
            sandbox_id=sandbox_id,
            volume_config=volume_config,
        )

        mount_request.additional_properties = d
        return mount_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="Toleration")


@_attrs_define
class Toleration:
    """
    Attributes:
        key (Union[Unset, str]):
        operator (Union[Unset, str]):
        value (Union[Unset, str]):
        effect (Union[Unset, str]):
    """

    key: Union[Unset, str] = UNSET
    operator: Union[Unset, str] = UNSET
    value: Union[Unset, str] = UNSET
    effect: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        key = self.key

        operator = self.operator

        value = self.value

        effect = self.effect

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if key is not UNSET:
            field_dict["key"] = key
        if operator is not UNSET:
            field_dict["operator"] = operator
        if value is not UNSET:
            field_dict["value"] = value
        if effect is not UNSET:
            field_dict["effect"] = effect

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        key = d.pop("key", UNSET)

        operator = d.pop("operator", UNSET)

        value = d.pop("value", UNSET)

        effect = d.pop("effect", UNSET)

        toleration = cls(
            key=key,
            operator=operator,
            value=value,
            effect=effect,
        )

        toleration.additional_properties = d
        return toleration

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.success_created_response import SuccessCreatedResponse
from ...models.success_written_response import SuccessWrittenResponse
from ...types import UNSET, File, Response, Unset


def _get_kwargs(
    id: str,
    *,
    body: File,
    path: str,
    mkdir: Union[Unset, bool] = UNSET,
    recursive: Union[Unset, bool] = UNSET,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    params: dict[str, Any] = {}

    params["path"] = path

    params["mkdir"] = mkdir

    params["recursive"] = recursive

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "post",
        "url": "/api/v1/sandboxes/{id}/files".format(
            id=id,
        ),
        "params": params,
    }

    _kwargs["content"] = body.payload

    headers["Content-Type"] = "application/octet-stream"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[SuccessCreatedResponse, SuccessWrittenResponse]]:
    if response.status_code == 200:
        response_200 = SuccessWrittenResponse.from_dict(response.json())

        return response_200

    if response.status_code == 201:
        response_201 = SuccessCreatedResponse.from_dict(response.json())

        return response_201

    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[SuccessCreatedResponse, SuccessWrittenResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    id: str,
    *,
    client: AuthenticatedClient,
    body: File,
    path: str,
    mkdir: Union[Unset, bool] = UNSET,
    recursive: Union[Unset, bool] = UNSET,
) -> Response[Union[SuccessCreatedResponse, SuccessWrittenResponse]]:
    """Write file or create directory

     Use `path` query param and `mkdir=true` to create directories, otherwise writes file content.

    Args:
        id (str):
        path (str):
        mkdir (Union[Unset, bool]):
        recursive (Union[Unset, bool]):
        body (File):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[SuccessCreatedResponse, SuccessWrittenResponse]]
    """

    kwargs = _get_kwargs(
        id=id,
        body=body,
        path=path,
        mkdir=mkdir,
        recursive=recursive,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    id: str,
    *,
    client: AuthenticatedClient,
    body: File,
    path: str,
    mkdir: Union[Unset, bool] = UNSET,
    recursive: Union[Unset, bool] = UNSET,
) -> Optional[Union[SuccessCreatedResponse, SuccessWrittenResponse]]:
    """Write file or create directory

     Use `path` query param and `mkdir=true` to create directories, otherwise writes file content.

    Args:
        id (str):
        path (str):
        mkdir (Union[Unset, bool]):
        recursive (Union[Unset, bool]):
        body (File):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[SuccessCreatedResponse, SuccessWrittenResponse]
    """

    return sync_detailed(
        id=id,
        client=client,
        body=body,
        path=path,
        mkdir=mkdir,
        recursive=recursive,
    ).parsed


async def asyncio_detailed(
    id: str,
    *,
    client: AuthenticatedClient,
    body: File,
    path: str,
    mkdir: Union[Unset, bool] = UNSET,
    recursive: Union[Unset, bool] = UNSET,
) -> Response[Union[SuccessCreatedResponse, SuccessWrittenResponse]]:
    """Write file or create directory

     Use `path` query param and `mkdir=true` to create directories, otherwise writes file content.

    Args:
        id (str):
        path (str):
        mkdir (Union[Unset, bool]):
        recursive (Union[Unset, bool]):
        body (File):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[SuccessCreatedResponse, SuccessWrittenResponse]]
    """

    kwargs = _get_kwargs(
        id=id,
        body=body,
        path=path,
        mkdir=mkdir,
        recursive=recursive,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    id: str,
    *,
    client: AuthenticatedClient,
    body: File,
    path: str,
    mkdir: Union[Unset, bool] = UNSET,
    recursive: Union[Unset, bool] = UNSET,
) -> Optional[Union[SuccessCreatedResponse, SuccessWrittenResponse]]:
    """Write file or create directory

     Use `path` query param and `mkdir=true` to create directories, otherwise writes file content.

    Args:
        id (str):
        path (str):
        mkdir (Union[Unset, bool]):
        recursive (Union[Unset, bool]):
        body (File):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[SuccessCreatedResponse, SuccessWrittenResponse]
    """

    return (
        await asyncio_detailed(
            id=id,
            client=client,
            body=body,
            path=path,
            mkdir=mkdir,
            recursive=recursive,
        )
    ).parsed



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional


@dataclass
class APIError(Exception):
    status_code: int
    message: str
    code: Optional[str] = None
    request_id: Optional[str] = None
    details: Any = None
    body: Optional[bytes] = None

    def __str__(self) -> str:
        parts = [f"status={self.status_code}"]
        if self.code:
            parts.append(f"code={self.code}")
        if self.request_id:
            parts.append(f"request_id={self.request_id}")
        parts.append(f"message={self.message}")
        return "APIError(" + ", ".join(parts) + ")"



from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.file_watch_subscribe_request_action import FileWatchSubscribeRequestAction
from ..types import UNSET, Unset

T = TypeVar("T", bound="FileWatchSubscribeRequest")


@_attrs_define
class FileWatchSubscribeRequest:
    """
    Attributes:
        action (FileWatchSubscribeRequestAction):
        path (str):
        recursive (Union[Unset, bool]):
    """

    action: FileWatchSubscribeRequestAction
    path: str
    recursive: Union[Unset, bool] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        action = self.action.value

        path = self.path

        recursive = self.recursive

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "action": action,
                "path": path,
            }
        )
        if recursive is not UNSET:
            field_dict["recursive"] = recursive

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        action = FileWatchSubscribeRequestAction(d.pop("action"))

        path = d.pop("path")

        recursive = d.pop("recursive", UNSET)

        file_watch_subscribe_request = cls(
            action=action,
            path=path,
            recursive=recursive,
        )

        file_watch_subscribe_request.additional_properties = d
        return file_watch_subscribe_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

from typing import TYPE_CHECKING

from sandbox0.apispec.api.sandbox_volumes import delete_api_v1_sandboxvolumes_id
from sandbox0.apispec.api.sandbox_volumes import get_api_v1_sandboxvolumes
from sandbox0.apispec.api.sandbox_volumes import get_api_v1_sandboxvolumes_id
from sandbox0.apispec.api.sandbox_volumes import post_api_v1_sandboxvolumes
from sandbox0.apispec.api.snapshots import delete_api_v1_sandboxvolumes_id_snapshots_snapshot_id
from sandbox0.apispec.api.snapshots import get_api_v1_sandboxvolumes_id_snapshots
from sandbox0.apispec.api.snapshots import get_api_v1_sandboxvolumes_id_snapshots_snapshot_id
from sandbox0.apispec.api.snapshots import post_api_v1_sandboxvolumes_id_snapshots
from sandbox0.apispec.api.snapshots import post_api_v1_sandboxvolumes_id_snapshots_snapshot_id_restore
from sandbox0.apispec.models.create_sandbox_volume_request import CreateSandboxVolumeRequest
from sandbox0.apispec.models.create_snapshot_request import CreateSnapshotRequest
from sandbox0.apispec.models.sandbox_volume import SandboxVolume
from sandbox0.apispec.models.snapshot import Snapshot
from sandbox0.apispec.models.success_deleted_response import SuccessDeletedResponse
from sandbox0.apispec.models.success_restore_response import SuccessRestoreResponse
from sandbox0.apispec.models.success_sandbox_volume_list_response import SuccessSandboxVolumeListResponse
from sandbox0.apispec.models.success_sandbox_volume_response import SuccessSandboxVolumeResponse
from sandbox0.apispec.models.success_snapshot_list_response import SuccessSnapshotListResponse
from sandbox0.apispec.models.success_snapshot_response import SuccessSnapshotResponse
from sandbox0.response import ensure_data, ensure_model

if TYPE_CHECKING:
    from sandbox0.client import Client


class ClientVolumesMixin:
    _api: any

    def create_volume(self: "Client", request: CreateSandboxVolumeRequest) -> SandboxVolume:
        resp = post_api_v1_sandboxvolumes.sync_detailed(client=self._api, body=request)
        return ensure_data(resp, SuccessSandboxVolumeResponse)

    def list_volumes(self: "Client") -> list[SandboxVolume]:
        resp = get_api_v1_sandboxvolumes.sync_detailed(client=self._api)
        data = ensure_data(resp, SuccessSandboxVolumeListResponse)
        return data

    # Backward-compatible alias.
    def list_volume(self: "Client") -> list[SandboxVolume]:
        return self.list_volumes()

    def get_volume(self: "Client", volume_id: str) -> SandboxVolume:
        resp = get_api_v1_sandboxvolumes_id.sync_detailed(id=volume_id, client=self._api)
        return ensure_data(resp, SuccessSandboxVolumeResponse)

    def delete_volume(self: "Client", volume_id: str) -> SuccessDeletedResponse:
        resp = delete_api_v1_sandboxvolumes_id.sync_detailed(id=volume_id, client=self._api)
        return ensure_model(resp, SuccessDeletedResponse)

    def create_volume_snapshot(self: "Client", volume_id: str, request: CreateSnapshotRequest) -> Snapshot:
        resp = post_api_v1_sandboxvolumes_id_snapshots.sync_detailed(id=volume_id, client=self._api, body=request)
        return ensure_data(resp, SuccessSnapshotResponse)

    def list_volume_snapshots(self: "Client", volume_id: str) -> list[Snapshot]:
        resp = get_api_v1_sandboxvolumes_id_snapshots.sync_detailed(id=volume_id, client=self._api)
        data = ensure_data(resp, SuccessSnapshotListResponse)
        return data

    def get_volume_snapshot(self: "Client", volume_id: str, snapshot_id: str) -> Snapshot:
        resp = get_api_v1_sandboxvolumes_id_snapshots_snapshot_id.sync_detailed(
            id=volume_id,
            snapshot_id=snapshot_id,
            client=self._api,
        )
        return ensure_data(resp, SuccessSnapshotResponse)

    def delete_volume_snapshot(self: "Client", volume_id: str, snapshot_id: str) -> SuccessDeletedResponse:
        resp = delete_api_v1_sandboxvolumes_id_snapshots_snapshot_id.sync_detailed(
            id=volume_id,
            snapshot_id=snapshot_id,
            client=self._api,
        )
        return ensure_model(resp, SuccessDeletedResponse)

    def restore_volume_snapshot(self: "Client", volume_id: str, snapshot_id: str) -> SuccessRestoreResponse:
        resp = post_api_v1_sandboxvolumes_id_snapshots_snapshot_id_restore.sync_detailed(
            id=volume_id,
            snapshot_id=snapshot_id,
            client=self._api,
        )
        return ensure_model(resp, SuccessRestoreResponse)



import datetime
from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.volume_access_mode import VolumeAccessMode
from ..types import UNSET, Unset

T = TypeVar("T", bound="SandboxVolume")


@_attrs_define
class SandboxVolume:
    """
    Attributes:
        id (str):
        team_id (str):
        user_id (str):
        cache_size (str):
        buffer_size (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        prefetch (Union[Unset, int]):
        writeback (Union[Unset, bool]):
        access_mode (Union[Unset, VolumeAccessMode]): Access mode for sandbox volumes. Enforcement is scoped to storage-
            proxy instances. RWO allows read-write mounts on a single instance; ROX allows read-only mounts across
            instances; RWX allows read-write mounts across instances.
    """

    id: str
    team_id: str
    user_id: str
    cache_size: str
    buffer_size: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    prefetch: Union[Unset, int] = UNSET
    writeback: Union[Unset, bool] = UNSET
    access_mode: Union[Unset, VolumeAccessMode] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        team_id = self.team_id

        user_id = self.user_id

        cache_size = self.cache_size

        buffer_size = self.buffer_size

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        prefetch = self.prefetch

        writeback = self.writeback

        access_mode: Union[Unset, str] = UNSET
        if not isinstance(self.access_mode, Unset):
            access_mode = self.access_mode.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "team_id": team_id,
                "user_id": user_id,
                "cache_size": cache_size,
                "buffer_size": buffer_size,
                "created_at": created_at,
                "updated_at": updated_at,
            }
        )
        if prefetch is not UNSET:
            field_dict["prefetch"] = prefetch
        if writeback is not UNSET:
            field_dict["writeback"] = writeback
        if access_mode is not UNSET:
            field_dict["access_mode"] = access_mode

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        id = d.pop("id")

        team_id = d.pop("team_id")

        user_id = d.pop("user_id")

        cache_size = d.pop("cache_size")

        buffer_size = d.pop("buffer_size")

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        prefetch = d.pop("prefetch", UNSET)

        writeback = d.pop("writeback", UNSET)

        _access_mode = d.pop("access_mode", UNSET)
        access_mode: Union[Unset, VolumeAccessMode]
        if isinstance(_access_mode, Unset):
            access_mode = UNSET
        else:
            access_mode = VolumeAccessMode(_access_mode)

        sandbox_volume = cls(
            id=id,
            team_id=team_id,
            user_id=user_id,
            cache_size=cache_size,
            buffer_size=buffer_size,
            created_at=created_at,
            updated_at=updated_at,
            prefetch=prefetch,
            writeback=writeback,
            access_mode=access_mode,
        )

        sandbox_volume.additional_properties = d
        return sandbox_volume

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import json
from typing import Any

import httpx

_NULL_MAP_KEYS = {
    "annotations",
    "envVars",
    "env_vars",
    "labels",
    "matchLabels",
    "nodeSelector",
}

_NULL_ARRAY_KEYS = {
    "allowedCidrs",
    "allowedDomains",
    "allowedPorts",
    "allowedTeams",
    "api_keys",
    "args",
    "candidates",
    "command",
    "conditions",
    "contexts",
    "data",
    "deniedCidrs",
    "deniedDomains",
    "deniedPorts",
    "drop",
    "entries",
    "env",
    "exposed_ports",
    "identities",
    "matchExpressions",
    "matchFields",
    "members",
    "mounts",
    "namespaces",
    "nodeSelectorTerms",
    "preferredDuringSchedulingIgnoredDuringExecution",
    "providers",
    "requiredDuringSchedulingIgnoredDuringExecution",
    "roles",
    "sidecars",
    "tags",
    "teams",
    "templates",
    "tolerations",
    "values",
}


def _normalize_null_string_map(value: dict[str, Any]) -> bool:
    changed = False
    for key, raw in value.items():
        if raw is None:
            value[key] = ""
            changed = True
    return changed


def _normalize_null_maps(payload: Any) -> tuple[Any, bool]:
    if isinstance(payload, dict):
        changed = False
        normalized: dict[str, Any] = {}
        for key, raw in payload.items():
            if raw is None and key in _NULL_MAP_KEYS:
                normalized[key] = {}
                changed = True
                continue
            if raw is None and key in _NULL_ARRAY_KEYS:
                normalized[key] = []
                changed = True
                continue
            if key in _NULL_MAP_KEYS and isinstance(raw, dict):
                changed = _normalize_null_string_map(raw) or changed
            next_value, nested_changed = _normalize_null_maps(raw)
            normalized[key] = next_value
            changed = nested_changed or changed
        return normalized, changed

    if isinstance(payload, list):
        changed = False
        normalized_list: list[Any] = []
        for item in payload:
            next_value, nested_changed = _normalize_null_maps(item)
            normalized_list.append(next_value)
            changed = nested_changed or changed
        return normalized_list, changed

    return payload, False


def normalize_response_json(response: httpx.Response) -> None:
    content_type = response.headers.get("content-type", "").lower()
    if "application/json" not in content_type:
        return

    body = response.content
    if not body:
        return

    try:
        payload = json.loads(body)
    except ValueError:
        return

    normalized, changed = _normalize_null_maps(payload)
    if not changed:
        return

    response._content = json.dumps(normalized, separators=(",", ":")).encode("utf-8")
    response.headers["content-length"] = str(len(response.content))


def normalize_response_hook(response: httpx.Response) -> None:
    response.read()
    normalize_response_json(response)


async def normalize_response_hook_async(response: httpx.Response) -> None:
    await response.aread()
    normalize_response_json(response)



from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="Snapshot")


@_attrs_define
class Snapshot:
    """
    Attributes:
        id (str):
        volume_id (str):
        name (str):
        size_bytes (int):
        created_at (str):
        description (Union[Unset, str]):
        expires_at (Union[None, Unset, str]):
    """

    id: str
    volume_id: str
    name: str
    size_bytes: int
    created_at: str
    description: Union[Unset, str] = UNSET
    expires_at: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        volume_id = self.volume_id

        name = self.name

        size_bytes = self.size_bytes

        created_at = self.created_at

        description = self.description

        expires_at: Union[None, Unset, str]
        if isinstance(self.expires_at, Unset):
            expires_at = UNSET
        else:
            expires_at = self.expires_at

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "volume_id": volume_id,
                "name": name,
                "size_bytes": size_bytes,
                "created_at": created_at,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description
        if expires_at is not UNSET:
            field_dict["expires_at"] = expires_at

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        id = d.pop("id")

        volume_id = d.pop("volume_id")

        name = d.pop("name")

        size_bytes = d.pop("size_bytes")

        created_at = d.pop("created_at")

        description = d.pop("description", UNSET)

        def _parse_expires_at(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        expires_at = _parse_expires_at(d.pop("expires_at", UNSET))

        snapshot = cls(
            id=id,
            volume_id=volume_id,
            name=name,
            size_bytes=size_bytes,
            created_at=created_at,
            description=description,
            expires_at=expires_at,
        )

        snapshot.additional_properties = d
        return snapshot

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.port_spec import PortSpec


T = TypeVar("T", bound="NetworkEgressPolicy")


@_attrs_define
class NetworkEgressPolicy:
    """
    Attributes:
        allowed_cidrs (Union[Unset, list[str]]):
        allowed_domains (Union[Unset, list[str]]):
        allowed_ports (Union[Unset, list['PortSpec']]):
        denied_domains (Union[Unset, list[str]]):
        denied_cidrs (Union[Unset, list[str]]):
        denied_ports (Union[Unset, list['PortSpec']]):
    """

    allowed_cidrs: Union[Unset, list[str]] = UNSET
    allowed_domains: Union[Unset, list[str]] = UNSET
    allowed_ports: Union[Unset, list["PortSpec"]] = UNSET
    denied_domains: Union[Unset, list[str]] = UNSET
    denied_cidrs: Union[Unset, list[str]] = UNSET
    denied_ports: Union[Unset, list["PortSpec"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        allowed_cidrs: Union[Unset, list[str]] = UNSET
        if not isinstance(self.allowed_cidrs, Unset):
            allowed_cidrs = self.allowed_cidrs

        allowed_domains: Union[Unset, list[str]] = UNSET
        if not isinstance(self.allowed_domains, Unset):
            allowed_domains = self.allowed_domains

        allowed_ports: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.allowed_ports, Unset):
            allowed_ports = []
            for allowed_ports_item_data in self.allowed_ports:
                allowed_ports_item = allowed_ports_item_data.to_dict()
                allowed_ports.append(allowed_ports_item)

        denied_domains: Union[Unset, list[str]] = UNSET
        if not isinstance(self.denied_domains, Unset):
            denied_domains = self.denied_domains

        denied_cidrs: Union[Unset, list[str]] = UNSET
        if not isinstance(self.denied_cidrs, Unset):
            denied_cidrs = self.denied_cidrs

        denied_ports: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.denied_ports, Unset):
            denied_ports = []
            for denied_ports_item_data in self.denied_ports:
                denied_ports_item = denied_ports_item_data.to_dict()
                denied_ports.append(denied_ports_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if allowed_cidrs is not UNSET:
            field_dict["allowedCidrs"] = allowed_cidrs
        if allowed_domains is not UNSET:
            field_dict["allowedDomains"] = allowed_domains
        if allowed_ports is not UNSET:
            field_dict["allowedPorts"] = allowed_ports
        if denied_domains is not UNSET:
            field_dict["deniedDomains"] = denied_domains
        if denied_cidrs is not UNSET:
            field_dict["deniedCidrs"] = denied_cidrs
        if denied_ports is not UNSET:
            field_dict["deniedPorts"] = denied_ports

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.port_spec import PortSpec

        d = dict(src_dict)
        allowed_cidrs = cast(list[str], d.pop("allowedCidrs", UNSET))

        allowed_domains = cast(list[str], d.pop("allowedDomains", UNSET))

        allowed_ports = []
        _allowed_ports = d.pop("allowedPorts", UNSET)
        for allowed_ports_item_data in _allowed_ports or []:
            allowed_ports_item = PortSpec.from_dict(allowed_ports_item_data)

            allowed_ports.append(allowed_ports_item)

        denied_domains = cast(list[str], d.pop("deniedDomains", UNSET))

        denied_cidrs = cast(list[str], d.pop("deniedCidrs", UNSET))

        denied_ports = []
        _denied_ports = d.pop("deniedPorts", UNSET)
        for denied_ports_item_data in _denied_ports or []:
            denied_ports_item = PortSpec.from_dict(denied_ports_item_data)

            denied_ports.append(denied_ports_item)

        network_egress_policy = cls(
            allowed_cidrs=allowed_cidrs,
            allowed_domains=allowed_domains,
            allowed_ports=allowed_ports,
            denied_domains=denied_domains,
            denied_cidrs=denied_cidrs,
            denied_ports=denied_ports,
        )

        network_egress_policy.additional_properties = d
        return network_egress_policy

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.label_selector import LabelSelector


T = TypeVar("T", bound="PodAffinityTerm")


@_attrs_define
class PodAffinityTerm:
    """
    Attributes:
        topology_key (str):
        label_selector (Union[Unset, LabelSelector]):
        namespaces (Union[Unset, list[str]]):
    """

    topology_key: str
    label_selector: Union[Unset, "LabelSelector"] = UNSET
    namespaces: Union[Unset, list[str]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        topology_key = self.topology_key

        label_selector: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.label_selector, Unset):
            label_selector = self.label_selector.to_dict()

        namespaces: Union[Unset, list[str]] = UNSET
        if not isinstance(self.namespaces, Unset):
            namespaces = self.namespaces

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "topologyKey": topology_key,
            }
        )
        if label_selector is not UNSET:
            field_dict["labelSelector"] = label_selector
        if namespaces is not UNSET:
            field_dict["namespaces"] = namespaces

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.label_selector import LabelSelector

        d = dict(src_dict)
        topology_key = d.pop("topologyKey")

        _label_selector = d.pop("labelSelector", UNSET)
        label_selector: Union[Unset, LabelSelector]
        if isinstance(_label_selector, Unset):
            label_selector = UNSET
        else:
            label_selector = LabelSelector.from_dict(_label_selector)

        namespaces = cast(list[str], d.pop("namespaces", UNSET))

        pod_affinity_term = cls(
            topology_key=topology_key,
            label_selector=label_selector,
            namespaces=namespaces,
        )

        pod_affinity_term.additional_properties = d
        return pod_affinity_term

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.user import User


T = TypeVar("T", bound="SuccessUserResponse")


@_attrs_define
class SuccessUserResponse:
    """
    Attributes:
        success (bool):
        data (Union[Unset, User]):
    """

    success: bool
    data: Union[Unset, "User"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        success = self.success

        data: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.data, Unset):
            data = self.data.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "success": success,
            }
        )
        if data is not UNSET:
            field_dict["data"] = data

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.user import User

        d = dict(src_dict)
        success = d.pop("success")

        _data = d.pop("data", UNSET)
        data: Union[Unset, User]
        if isinstance(_data, Unset):
            data = UNSET
        else:
            data = User.from_dict(_data)

        success_user_response = cls(
            success=success,
            data=data,
        )

        success_user_response.additional_properties = d
        return success_user_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
